/************************************************************************/
/*  Ted: Interface to the find tool.					*/
/************************************************************************/

#   include	"tedConfig.h"

#   include	<stddef.h>
#   include	<stdlib.h>
#   include	<stdio.h>
#   include	<ctype.h>

#   include	"tedApp.h"

#   include	<appSpellTool.h>
#   include	<docFind.h>
#   include	<ind.h>

#   include	<appDebugon.h>

static void tedFindToolClosed(		void *		voidea )
    {
    EditApplication *	ea=	(EditApplication *)voidea;

    ea->eaFindTool= (void *)0;
    }

int tedDocFindNext(	EditDocument *	ed )
    {
    TedDocument *		td;
    BufferDocument *		bd;

    int				ret;

    DocumentSelection		dsNew;

    int				scrolledX= 0;
    int				scrolledY= 0;

    DocumentSelection		ds;
    SelectionGeometry		sg;
    SelectionDescription	sd;

    td= (TedDocument *)ed->edPrivateData;
    bd= td->tdDocument;

    if  ( ! td->tdFindProg )
	{ XDEB(td->tdFindProg); return -1;	}

    if  ( tedGetSelection( &ds, &sg, &sd,
			    (DocumentTree **)0, (const BufferItem **)0, td ) )
	{ LDEB(1); return -1;	}

    docInitDocumentSelection( &dsNew );

    ret= docFindFindNextInDocument( &dsNew, &(ds.dsTail), bd,
				docFindParaFindNext, (void *)td->tdFindProg );

    if  ( ret < 0 )
	{ LDEB(ret); return -1;	}

    if  ( ret == 0 )
	{
	EditApplication *	ea= ed->edApplication;
	const int		lastLine= 0;

	tedSetSelection( ed, &dsNew, lastLine, &scrolledX, &scrolledY );

	tedAdaptToolsToSelection( ed );

	if  ( ea->eaFindTool )
	    { appFindToolEnableReplace( ea->eaFindTool, 1 );	}
	}

    return ret;
    }

int tedDocFindPrev(	EditDocument *	ed )
    {
    TedDocument *		td;
    BufferDocument *		bd;

    int				ret;

    DocumentSelection		dsNew;

    int				scrolledX= 0;
    int				scrolledY= 0;

    DocumentSelection		ds;
    SelectionGeometry		sg;
    SelectionDescription	sd;

    td= (TedDocument *)ed->edPrivateData;
    bd= td->tdDocument;

    if  ( ! td->tdFindProg )
	{ XDEB(td->tdFindProg); return -1;	}

    if  ( tedGetSelection( &ds, &sg, &sd,
			    (DocumentTree **)0, (const BufferItem **)0, td ) )
	{ LDEB(1); return -1;	}

    docInitDocumentSelection( &dsNew );

    ret= docFindFindPrevInDocument( &dsNew, &(ds.dsHead), bd,
				docFindParaFindPrev, (void *)td->tdFindProg );

    if  ( ret < 0 )
	{ LDEB(ret); return -1;	}

    if  ( ret == 0 )
	{
	EditApplication *	ea= ed->edApplication;
	const int		lastLine= 0;

	tedSetSelection( ed, &dsNew, lastLine, &scrolledX, &scrolledY );

	tedAdaptToolsToSelection( ed );

	if  ( ea->eaFindTool )
	    { appFindToolEnableReplace( ea->eaFindTool, 1 );	}
	}

    return ret;
    }

static int tedFindToolFindNext(		void *		voidea )
    {
    EditApplication *		ea= (EditApplication *)voidea;
    EditDocument *		ed= ea->eaCurrentDocument;

    if  ( ! ed )
	{ XDEB(ed); return -1;	}

    return tedDocFindNext( ed );
    }

static int tedFindToolFindPrev(		void *		voidea )
    {
    EditApplication *		ea= (EditApplication *)voidea;
    EditDocument *		ed= ea->eaCurrentDocument;

    if  ( ! ed )
	{ XDEB(ed); return -1;	}

    return tedDocFindPrev( ed );
    }

static int tedFindToolSetPattern(	void *		voidea,
					const char *	pattern,
					int		useRegex )
    {
    EditApplication *		ea= (EditApplication *)voidea;
    EditDocument *		ed= ea->eaCurrentDocument;
    TedDocument *		td;

    if  ( ! ed )
	{ XDEB(ed); return -1;	}

    td= (TedDocument *)ed->edPrivateData;

    return docFindSetPattern( &(td->tdFindProg), pattern, useRegex );
    }

static void tedShowFindTool(	EditDocument *		ed )
    {
    EditApplication *	ea= ed->edApplication;

    if  ( ! ea->eaFindTool )
	{
	const char *		pixmapName= "tedfind";
	APP_BITMAP_IMAGE	iconPixmap= (APP_BITMAP_IMAGE)0;
	APP_BITMAP_MASK		iconMask= (APP_BITMAP_MASK)0;
	const TedDocument *	td= (TedDocument *)ed->edPrivateData;

	TedAppResources *	tar= (TedAppResources *)ea->eaResourceData;

	if  ( appGetImagePixmap( ea, pixmapName, &iconPixmap, &iconMask )  )
	    { SDEB(pixmapName); return;	}

	ea->eaFindTool= appMakeFindTool( td->tdFindOption,
			    ea, "tedFindTool",
			    iconPixmap, iconMask,
			    tedFindToolClosed,
			    tedFindToolFindNext,
			    tedFindToolFindPrev,
			    tedAppReplace,
			    tedFindToolSetPattern,
			    (void *)ea );
	if  ( ! ea->eaFindTool )
	    { XDEB(ea->eaFindTool); return;	}

	if  ( tar->tarFindPattern )
	    {
	    appFindToolSetPattern( ea->eaFindTool,
				    tar->tarFindPattern, tar->tarFindRegex );
	    }
	}

    appShowFindTool( ed->edToplevel.atTopWidget, ea->eaFindTool );

    if  ( ea->eaFindTool )
	{ appFindToolEnableReplace( ea->eaFindTool, 1 );	}

    return;
    }

APP_MENU_CALLBACK_H( tedDocToolFind, findOption, voided, e )
    {
    EditDocument *	ed= (EditDocument *)voided;

    tedShowFindTool( ed );

    return;
    }

APP_MENU_CALLBACK_H( tedDocToolFindNext, findOption, voided, e )
    {
    EditDocument *	ed= (EditDocument *)voided;
    TedDocument *	td;

    if  ( ! ed )
	{ XDEB(ed); return;	}

    td= (TedDocument *)ed->edPrivateData;

    if  ( ! td->tdFindProg )
	{
	tedShowFindTool( ed );
	return;
	}

    tedDocFindNext( ed );

    return;
    }

static int tedSpellFindNext(	void *			voidea,
				const SpellChecker *	spc,
				SpellDictionary *	spd )
    {
    EditApplication *		ea= (EditApplication *)voidea;
    EditDocument *		ed= ea->eaCurrentDocument;
    TedDocument *		td;
    BufferDocument *		bd;

    int				ret;

    DocumentSelection		dsNew;

    int				scrolledX= 0;
    int				scrolledY= 0;

    DocumentSelection		ds;
    SelectionGeometry		sg;
    SelectionDescription	sd;

    if  ( ! ed )
	{ XDEB(ed); return -1;	}

    td= (TedDocument *)ed->edPrivateData;
    bd= td->tdDocument;

    if  ( tedGetSelection( &ds, &sg, &sd,
			    (DocumentTree **)0, (const BufferItem **)0, td ) )
	{ LDEB(1); return -1;	}

    docInitDocumentSelection( &dsNew );

    ret= docFindFindNextInDocument( &dsNew, &(ds.dsTail), bd,
					docSpellParaFindNext, (void *)spd );

    if  ( ret < 0 )
	{ LDEB(ret); return -1;	}

    if  ( ret == 0 )
	{
	const int	lastLine= 0;

	tedSetSelection( ed, &dsNew, lastLine, &scrolledX, &scrolledY );

	tedAdaptToolsToSelection( ed );

	if  ( ea->eaSpellTool )
	    {
	    MemoryBuffer	mbGuess;

	    utilInitMemoryBuffer( &mbGuess );

	    if  ( utilMemoryBufferGetRange(
			&mbGuess, &(dsNew.dsTail.dpBi->biParaStringBuffer),
			dsNew.dsHead.dpStroff,
			dsNew.dsTail.dpStroff- dsNew.dsHead.dpStroff ) )
		{ LDEB(1); return -1;	}

	    appSpellMakeGuesses( ea->eaSpellTool, (char *)mbGuess.mbBytes );

	    utilCleanMemoryBuffer( &mbGuess );
	    }
	}

    return ret;
    }

/************************************************************************/
/*									*/
/*  Show or create a spelling tool.					*/
/*									*/
/************************************************************************/

static SpellChecker TED_SpellChecker;

static void tedSpellToolClosed(		void *		voidea	)
    {
    EditApplication *	ea= (EditApplication *)voidea;

    ea->eaSpellTool= (void *)0;

    indCleanSpellChecker( &TED_SpellChecker );
    indInitSpellChecker( &TED_SpellChecker );
    }

APP_MENU_CALLBACK_H( tedDocToolSpell, spellOption, voided, e )
    {
    EditDocument *	ed=	(EditDocument *)voided;
    EditApplication *	ea=	ed->edApplication;

    if  ( ! ea->eaSpellTool )
	{
	const char *		pixmapName= "tedspell";
	APP_BITMAP_IMAGE	iconPixmap= (APP_BITMAP_IMAGE)0;
	APP_BITMAP_MASK		iconMask= (APP_BITMAP_MASK)0;

	if  ( appGetImagePixmap( ea, pixmapName, &iconPixmap, &iconMask )  )
	    { SDEB(pixmapName); return;	}

	TED_SpellChecker.scSetup= indSpellSetup;
	TED_SpellChecker.scFindNext= tedSpellFindNext;
	TED_SpellChecker.scCorrect= tedAppReplace;

	ea->eaSpellTool= appMakeSpellTool( spellOption, ea, "tedSpellTool",
			iconPixmap, iconMask,
			tedSpellToolClosed, &TED_SpellChecker, (void *)ea );
	}
    else{ appShowSpellTool( ea->eaSpellTool ); }

    return;
    }
