/************************************************************************/
/*									*/
/*  Drawing primitives for X11.						*/
/*									*/
/************************************************************************/

#   include	"appFrameConfig.h"

#   include	<stddef.h>
#   include	<stdio.h>
#   include	<limits.h>
#   include	<math.h>

#   include	<geo2DInteger.h>
#   include	<uniUtf8.h>
#   include	"appDraw.h"
#   include	"appScreenFont.h"

#   include	<appDebugon.h>

#   ifdef USE_X11_FONTS

static XChar2b *	ws;
static int		wc= 0;

static unsigned char *	bs;
static int		bc= 0;

typedef struct TextProgress
    {
    int			tpX;
    int			tpY;
    } TextProgress;

typedef void (*HandleText8)(	TextProgress *		tp,
				AppDrawingData *	add,
				APP_FONT *		xfs,
				const char *		s,
				int			l );

typedef void (*HandleText16)(	TextProgress *		tp,
				AppDrawingData *	add,
				APP_FONT *		xfs,
				const XChar2b *		ws,
				int			wl );

/************************************************************************/
/*									*/
/*  Actual actions.							*/
/*									*/
/************************************************************************/

#   ifdef USE_GTK

static void appX11DrawTextSegment8(	TextProgress *		tp,
					AppDrawingData *	add,
					APP_FONT *		xfs,
					const char *		s,
					int			l )
    {
    gdk_draw_text( add->addDrawable, xfs, add->addGc,
					tp->tpX, tp->tpY, s, l );
    }

static void appX11DrawTextSegment16(	TextProgress *		tp,
					AppDrawingData *	add,
					APP_FONT *		xfs,
					const XChar2b *		ws,
					int			wl )
    {
    gdk_draw_text( add->addDrawable, xfs, add->addGc,
					tp->tpX, tp->tpY, (gchar *)ws, 2* wl );
    }

static void appX11TextSegmentLength8(	TextProgress *		tp,
					AppDrawingData *	add,
					APP_FONT *		xfs,
					const char *		s,
					int			l )
    {
    tp->tpX += gdk_text_width( xfs, s, l );
    }

static void appX11TextSegmentLength16(	TextProgress *		tp,
					AppDrawingData *	add,
					APP_FONT *		xfs,
					const XChar2b *		ws,
					int			wl )
    {
    tp->tpX += gdk_text_width( xfs, (gchar *)ws, 2* wl );
    }

#   endif

/************************************************************************/
/*									*/
/*  Actual actions.							*/
/*									*/
/************************************************************************/

#   ifdef USE_MOTIF

static void appX11DrawTextSegment8(	TextProgress *		tp,
					AppDrawingData *	add,
					APP_FONT *		xfs,
					const char *		s,
					int			l )
    {
    if  ( add->addCurrentFID != xfs->fid )
	{
	XSetFont( add->addDisplay, add->addGc, xfs->fid );
	add->addCurrentFID= xfs->fid;
	}

    XDrawString( add->addDisplay, add->addDrawable, add->addGc,
						    tp->tpX, tp->tpY, s, l );
    }

static void appX11DrawTextSegment16(	TextProgress *		tp,
					AppDrawingData *	add,
					APP_FONT *		xfs,
					const XChar2b *		ws,
					int			wl )
    {
    if  ( add->addCurrentFID != xfs->fid )
	{
	XSetFont( add->addDisplay, add->addGc, xfs->fid );
	add->addCurrentFID= xfs->fid;
	}

    XDrawString16( add->addDisplay, add->addDrawable, add->addGc,
						tp->tpX, tp->tpY, ws, wl );
    }

static void appX11TextSegmentLength8(	TextProgress *		tp,
					AppDrawingData *	add,
					APP_FONT *		xfs,
					const char *		s,
					int			l )
    {
    tp->tpX += XTextWidth( xfs, s, l );
    }

static void appX11TextSegmentLength16(	TextProgress *		tp,
					AppDrawingData *	add,
					APP_FONT *		xfs,
					const XChar2b *		ws,
					int			wl )
    {
    tp->tpX += XTextWidth16( xfs, ws, wl );
    }

#   endif

/************************************************************************/
/*									*/
/*  Convert to 16 bit string.						*/
/*									*/
/************************************************************************/

static int appDrawX11SetWS(	int *			pUsed,
				const char *		ss,
				int			len )
    {
    const unsigned char *	us= (const unsigned char *)ss;
    int				wl= 0;
    int				used= 0;

    if  ( wc < len )
	{
	XChar2b *	fresh= realloc( ws, len* sizeof( XChar2b ) );

	if  ( ! fresh )
	    { LXDEB(len,fresh);	return -1;	}

	ws= fresh; wc= len;
	}

    while( len > 0 )
	{
	unsigned short	sym;
	int		step;

	step= uniGetUtf8( &sym, us );
	if  ( step < 1 )
	    { LDEB(step); return -1;	}

	ws[wl].byte1= sym/ 256; ws[wl].byte2= sym% 256; wl++;
	used += step; us += step; len -= step;
	}

    *pUsed= used; return wl;
    }

/************************************************************************/
/*									*/
/*  Convert to 8 bit string.						*/
/*									*/
/************************************************************************/

static int appDrawX11SetBS(	int *			pUsed,
				const IndexMapping *	im,
				const char *		ss,
				int			len )
    {
    const unsigned char *	us= (const unsigned char *)ss;
    int				bl= 0;
    int				used= 0;

    if  ( ! im )
	{ XDEB(im); return -1;	}

    if  ( bc < len )
	{
	unsigned char *	fresh= realloc( bs, len* sizeof( unsigned char ) );

	if  ( ! fresh )
	    { LXDEB(len,fresh);	return -1;	}

	bs= fresh; bc= len;
	}

    while( len > 0 )
	{
	unsigned short	sym;
	int		step;
	int		byte;

	step= uniGetUtf8( &sym, us );
	if  ( step < 1 )
	    { LDEB(step); return -1;	}

	byte= utilIndexMappingGetU( im, sym );
	if  ( byte < 0 )
	    { break;	}

	bs[bl]= byte; bl++;
	used += step; us += step; len -= step;
	}

    *pUsed= used; return bl;
    }

void appDrawSetFont(		AppDrawingData *	add,
				int			screenFont )
    {
    const PagedList *	apfl= &(add->addScreenFontList);
    DrawScreenFont *	dsf;

    dsf= (DrawScreenFont *)utilPagedListGetItemByNumber( apfl, screenFont );
    if  ( ! dsf )
	{ LXDEB(screenFont,dsf); return;	}

    if  ( add->addCurrentScreenFontNumber != screenFont )
	{
	add->addCurrentScreenFontNumber= screenFont;
#	ifdef USE_MOTIF
	add->addCurrentFID= None;
#	endif
	}

    return;
    }

static void appDrawX11HandleText( TextProgress *		tp,
				AppDrawingData *		add,
				const char *			s,
				int				len,
				HandleText8			handleText8,
				HandleText16			handleText16,
				const EncodedScreenFontList *	esfl )
    {
    int				used= 0;

    if  ( esfl->esfFontCount == 0 )
	{ LDEB(esfl->esfFontCount); return;	}

    while( len > 0 )
	{
	const EncodedScreenFont *	efs;
	int				font;
	int				step;
	unsigned short			sym;

	step= uniGetUtf8( &sym, (unsigned char *)s );
	if  ( step < 1 )
	    { LLDEB(len,step); return;	}

	font= utilIndexMappingGetU( &(esfl->esflSymbolToFont), sym );
	if  ( font < 0 )
	    { /*XLDEB(sym,font);*/ len -= step; s += step; continue;	}
	efs= esfl->esfFonts+ font;

	if  ( efs->esfIsTwoByte )
	    {
	    int	wl= appDrawX11SetWS( &used, s, len );

	    if  ( wl < 0 )
		{ LLDEB(len,wl); return;	}

	    (*handleText16)( tp, add, efs->esfFontStruct, ws, wl );
	    }
	else{
	    const IndexMapping *	s2b= efs->esfSymbolToByteMapping;
	    int				bl;

	    bl= appDrawX11SetBS( &used, s2b, s, len );
	    if  ( bl < 0 )
		{ SLLDEB(efs->esfFontName,len,bl); return;	}

	    (*handleText8)( tp, add, efs->esfFontStruct, (char *)bs, bl );
	    }

	if  ( used < 1 )
	    { LDEB(used); return;	}

	len -= used; s += used;
	}

    return;
    }

void appDrawDrawString(		AppDrawingData *	add,
				int			x,
				int			y,
				const char *		s,
				int			len )
    {
    const PagedList *	apfl= &(add->addScreenFontList);
    DrawScreenFont *	dsf;

    TextProgress	tp;

    dsf= (DrawScreenFont *)utilPagedListGetItemByNumber( apfl,
					    add->addCurrentScreenFontNumber );
    if  ( ! dsf )
	{ LXDEB(add->addCurrentScreenFontNumber,dsf); return;	}

#   ifdef USE_XFT
    if  ( dsf->apfXftFont						    &&
	  ! appDrawDrawStringXft( add, dsf, dsf->apfXftFont, x, y, s, len ) )
	{ return;	}
#   endif

    tp.tpX= x;
    tp.tpY= y;

    appDrawX11HandleText( &tp, add, s, len, appX11DrawTextSegment8,
			appX11DrawTextSegment16, &(dsf->apfEncodedFonts) );
    }

void appDrawDrawScapsString(	AppDrawingData *	add,
				int			x,
				int			y,
				const char *		s,
				int			len )
    {
    const PagedList *	apfl= &(add->addScreenFontList);
    DrawScreenFont *	dsf;

    TextProgress	tp;

    dsf= (DrawScreenFont *)utilPagedListGetItemByNumber( apfl,
					    add->addCurrentScreenFontNumber );
    if  ( ! dsf )
	{ LXDEB(add->addCurrentScreenFontNumber,dsf); return;	}

#   ifdef USE_XFT
    if  ( dsf->apfXftScapsFont						&&
	  ! appDrawDrawStringXft( add, dsf, dsf->apfXftScapsFont,
							x, y, s, len )	)
	{ return;	}
#   endif

    tp.tpX= x;
    tp.tpY= y;

    appDrawX11HandleText( &tp, add, s, len,
		    appX11DrawTextSegment8, appX11DrawTextSegment16,
		    &(dsf->apfScapsEncodedFonts) );
    }

/************************************************************************/
/*									*/
/*  Determine text extents. NOTE that in the scaps case, the regular	*/
/*  font is used for the vertical dimensions and the scaps variant for	*/
/*  the horizontal dimensions.						*/
/*									*/
/************************************************************************/

static void appDrawVerticalExtents(
				DocumentRectangle *		drText,
				int				y,
				const EncodedScreenFontList *	esfl )
    {
    const EncodedScreenFont *	esf;
    int				i;
    int				ascent= 0;
    int				descent= 0;

    esf= esfl->esfFonts;
    for ( i= 0; i < esfl->esfFontCount; esf++, i++ )
	{
	if  ( ascent < esf->esfFontStruct->ascent )
	    { ascent=  esf->esfFontStruct->ascent;	}
	if  ( descent < esf->esfFontStruct->descent )
	    { descent=  esf->esfFontStruct->descent;	}
	}

    drText->drY0= y- descent;
    drText->drY1= y- ascent;
    }

void appDrawTextExtents(	DocumentRectangle *	drText,
				int			x,
				int			y,
				AppDrawingData *	add,
				int			screenFont,
				const char *		s,
				int			len )
    {
    const PagedList *	apfl= &(add->addScreenFontList);
    DrawScreenFont *	dsf;

    TextProgress	tp;

    dsf= (DrawScreenFont *)utilPagedListGetItemByNumber( apfl, screenFont );
    if  ( ! dsf )
	{ LXDEB(screenFont,dsf); return;	}

#   ifdef USE_XFT
    if  ( dsf->apfXftFont						&&
	  ! appDrawTextExtentsXft( drText, x, y,
			add, dsf, dsf->apfXftFont, dsf->apfXftFont, s, len ) )
	{ return;	}
#   endif

    tp.tpX= x;
    tp.tpY= y;

    appDrawX11HandleText( &tp, add, s, len,
		    appX11TextSegmentLength8, appX11TextSegmentLength16,
		    &(dsf->apfEncodedFonts) );

    drText->drX0= x;
    drText->drX1= tp.tpX;

    appDrawVerticalExtents( drText, y, &(dsf->apfEncodedFonts) );

    return;
    }

void appDrawScapsTextExtents(	DocumentRectangle *	drText,
				int			x,
				int			y,
				AppDrawingData *	add,
				int			screenFont,
				const char *		s,
				int			len )
    {
    const PagedList *	apfl= &(add->addScreenFontList);
    DrawScreenFont *	dsf;

    TextProgress	tp;

    dsf= (DrawScreenFont *)utilPagedListGetItemByNumber( apfl, screenFont );
    if  ( ! dsf )
	{ LXDEB(screenFont,dsf); return;	}

#   ifdef USE_XFT
    if  ( dsf->apfXftFont						&&
	  dsf->apfXftScapsFont						&&
	  ! appDrawTextExtentsXft( drText, x, y,
		    add, dsf, dsf->apfXftScapsFont, dsf->apfXftFont, s, len ) )
	{ return;	}
#   endif

    tp.tpX= x;
    tp.tpY= y;

    appDrawX11HandleText( &tp, add, s, len,
		    appX11TextSegmentLength8, appX11TextSegmentLength16,
		    &(dsf->apfScapsEncodedFonts) );

    drText->drX0= x;
    drText->drX1= tp.tpX;

    /* note: no scaps here */
    appDrawVerticalExtents( drText, y, &(dsf->apfEncodedFonts) );

    return;
    }

#   endif
