/************************************************************************/
/*									*/
/*  Basic screen font management.					*/
/*									*/
/************************************************************************/

#   include	"appFrameConfig.h"

#   include	<stddef.h>
#   include	<stdlib.h>
#   include	<stdio.h>

#   include	<geo2DInteger.h>
#   include	<psFontMetrics.h>
#   include	"appDraw.h"
#   include	"appScreenFont.h"
#   include	"appMatchFont.h"

#   include	<appDebugon.h>

void appDrawUnderlineGeometry(		int *			pThick,
					int *			pY,
					const AppDrawingData *	add,
					int			screenFont,
					int			baselinePixels )
    {
    const PagedList *		apfl= &(add->addScreenFontList);
    const DrawScreenFont *	dsf;

    dsf= (DrawScreenFont *)utilPagedListGetItemByNumber( apfl, screenFont );
    if  ( ! dsf )
	{  LXDEB(screenFont,dsf); return;	}

    *pThick= dsf->apfUnderlineThicknessPixels;
    *pY= baselinePixels+ dsf->apfUnderlinePositionPixels;
    return;
    }

void appDrawStrikethroughGeometry(	int *			pThick,
					int *			pY,
					const AppDrawingData *	add,
					int			screenFont,
					int			baselinePixels )
    {
    const PagedList *		apfl= &(add->addScreenFontList);
    const DrawScreenFont *	dsf;

    dsf= (DrawScreenFont *)utilPagedListGetItemByNumber( apfl, screenFont );
    if  ( ! dsf )
	{  LXDEB(screenFont,dsf); return;	}

    *pThick= dsf->apfStrikethroughThicknessPixels;
    *pY= baselinePixels+ dsf->apfStrikethroughPositionPixels;
    return;
    }

void appDrawGetSuperBaseline(	int *			pSuperBaseline,
				int			baselinePixels,
				const AppDrawingData *	add,
				int			screenFont )
    {
    const PagedList *		apfl= &(add->addScreenFontList);
    const DrawScreenFont *	dsf;

    dsf= (DrawScreenFont *)utilPagedListGetItemByNumber( apfl, screenFont );
    if  ( ! dsf )
	{  LXDEB(screenFont,dsf); return;	}

    *pSuperBaseline= baselinePixels+ dsf->apfSuperBaseline;
    return;
    }

void appDrawGetSubBaseline(	int *			pSubBaseline,
				int			baselinePixels,
				const AppDrawingData *	add,
				int			screenFont )
    {
    const PagedList *		apfl= &(add->addScreenFontList);
    const DrawScreenFont *	dsf;

    dsf= (DrawScreenFont *)utilPagedListGetItemByNumber( apfl, screenFont );
    if  ( ! dsf )
	{  LXDEB(screenFont,dsf); return;	}

    *pSubBaseline= baselinePixels+ dsf->apfSubBaseline;
    return;
    }

void appDrawCleanEncodedScreenFont(	DrawScreenFont *	dsf,
					EncodedScreenFont *	esf )
    {
    if  ( esf->esfFontStruct )
	{
#	ifdef USE_MOTIF
	if  ( ! esf->esfFontBorrowed )
	    { XFreeFont( dsf->apfDisplay, esf->esfFontStruct );	}
#	endif

#	ifdef USE_GTK
	gdk_font_unref( esf->esfFontStruct );
#	endif
	}

    return;
    }

static void appCleanEncodedScreenFontList(	DrawScreenFont *	dsf,
						EncodedScreenFontList * esfl )
    {
    int			i;

    for ( i= 0; i < esfl->esfFontCount; i++ )
	{ appDrawCleanEncodedScreenFont( dsf, esfl->esfFonts+ i ); }

    if  ( esfl->esfFonts )
	{ free( esfl->esfFonts );	}

    utilCleanIndexMapping( &(esfl->esflSymbolToFont) );
    }

static void appDrawCleanScreenFont(	void *	vdsf )
    {
    DrawScreenFont *	dsf= (DrawScreenFont *)vdsf;

#   ifdef USE_XFT
    if  ( dsf->apfXftFont )
	{ XftFontClose( dsf->apfDisplay, dsf->apfXftFont );	}

    if  ( dsf->apfXftScapsFont )
	{ XftFontClose( dsf->apfDisplay, dsf->apfXftScapsFont );	}
#   endif

    appCleanEncodedScreenFontList( dsf, &(dsf->apfEncodedFonts) );
    appCleanEncodedScreenFontList( dsf, &(dsf->apfScapsEncodedFonts) );

    return;
    }

static void appInitEncodedScreenFontList(	EncodedScreenFontList * esfl )
    {
    esfl->esfFonts= (EncodedScreenFont *)0;
    esfl->esfFontCount= 0;

    utilInitIndexMapping( &(esfl->esflSymbolToFont) );
    }

static void appDrawInitScreenFont(	void *	vdsf )
    {
    DrawScreenFont *	dsf= (DrawScreenFont *)vdsf;

    /*  KEY  */
    dsf->apfPsFaceNumber= -1;
    dsf->apfFullSizePixels= 0;

    /*  DEP  */
    dsf->apfPsFontInfo= (AfmFontInfo *)0;
    dsf->apfUnderlinePositionPixels= 0;
    dsf->apfUnderlineThicknessPixels= 0;
    dsf->apfStrikethroughPositionPixels= 0;
    dsf->apfStrikethroughThicknessPixels= 0;
    dsf->apfSuperBaseline= 0;
    dsf->apfSubBaseline= 0;
    dsf->apfXHeightPixels= 0;
    dsf->apfScapsSizePixels= 0;

    appInitEncodedScreenFontList( &(dsf->apfEncodedFonts) );
    appInitEncodedScreenFontList( &(dsf->apfScapsEncodedFonts) );

    dsf->apfFontName[0]= '\0';

    dsf->apfUnicodesUsed= (const IndexSet *)0;

#   if defined(USE_XFT) || defined(USE_MOTIF)
    dsf->apfDisplay= (Display *)0;
#   endif

#   ifdef USE_XFT
    dsf->apfXftFont= (XftFont *)0;
    dsf->apfXftScapsFont= (XftFont *)0;
#   endif

    return;
    }

void appDrawCleanScreenFontList(	AppDrawingData *	add )
    {
    utilCleanPagedList( &(add->addScreenFontList) );

    return;
    }

void appDrawStartScreenFontList(	AppDrawingData *	add )
    {
    utilInitPagedList( &(add->addScreenFontList) );

    utilStartPagedList( &(add->addScreenFontList), sizeof(DrawScreenFont),
			    appDrawInitScreenFont, appDrawCleanScreenFont );

    return;
    }

/************************************************************************/
/*									*/
/*  Find the slot for a particular font/attribute combination.		*/
/*									*/
/*  1)  Is it an open font? If so return its number.			*/
/*  2)  Claim memory.							*/
/*  3)  Obtain a list of the available fonts.				*/
/*									*/
/************************************************************************/

static int appGetScreenFont(	DrawScreenFont **		pApf,
				int *				pFresh,
				AppDrawingData *		add,
				DocumentFontList *		dfl,
				const TextAttribute *		ta )
    {
    PagedList *				apfl= &(add->addScreenFontList);
    const PostScriptFontList *		psfl= add->addPostScriptFontList;

    DocumentFont *			df;
    const int				twipsSize= 10* ta->taFontSizeHalfPoints;

    int					screenFont;

    int					fullSizePixels;
    const AfmFontInfo *			afi;
    DrawScreenFont *			dsf;

    df= docFontListGetFontByNumber( dfl, ta->taFontNumber );
    if  ( ! df )
	{ LXDEB(ta->taFontNumber,df); return -1;	}

    afi= appGetFontInfoForAttribute( ta, dfl, psfl );
    if  ( ! afi )
	{ SXDEB(df->dfName,afi); return -1;	}

    if  ( ta->taSuperSub == DOCfontSUPERSCRIPT		||
	  ta->taSuperSub == DOCfontSUBSCRIPT		)
	{
	fullSizePixels= COORDtoGRID( add->addMagnifiedPixelsPerTwip,
						SUPERSUB_SIZE( twipsSize ) );
	}
    else{
	fullSizePixels= COORDtoGRID( add->addMagnifiedPixelsPerTwip,
								twipsSize );
	}

    if  ( fullSizePixels < 2 )
	{
	/* LFLDEB(twipsSize,add->addMagnifiedPixelsPerTwip,fullSizePixels); */
	fullSizePixels= 2;
	}


    /*  1  */
    for ( screenFont= 0; screenFont < apfl->plItemCount; screenFont++ )
	{
	dsf= utilPagedListGetItemByNumber( apfl, screenFont );
	if  ( ! dsf )
	    { LXDEB(screenFont,dsf); continue;	}

	if  ( afi->afiFaceNumber == dsf->apfPsFaceNumber	&&
	      fullSizePixels == dsf->apfFullSizePixels		)
	    { *pApf= dsf; *pFresh= 0; return screenFont;	}
	}

    /*  2  */
    dsf= utilPagedListClaimItemAtEnd( &screenFont, apfl );
    if  ( ! dsf )
	{ XDEB(dsf); return -1;	}

    /*  KEY  */
    dsf->apfPsFaceNumber= afi->afiFaceNumber;
    dsf->apfFullSizePixels= fullSizePixels;

    /*  DEP  */
    dsf->apfPsFontInfo= afi;
    dsf->apfUnicodesUsed= &(df->dfUnicodesUsed);

    if  ( ta->taSuperSub == DOCfontSUPERSCRIPT		||
	  ta->taSuperSub == DOCfontSUBSCRIPT		)
	{
	dsf->apfScapsSizePixels= COORDtoGRID( add->addMagnifiedPixelsPerTwip,
						SUPERSUB_SIZE( twipsSize ) );
	}
    else{
	int		scapsSize;

	psGetSmallcapsSize( &scapsSize, twipsSize, afi );
	dsf->apfScapsSizePixels= COORDtoGRID( add->addMagnifiedPixelsPerTwip,
								    scapsSize );
	}

    if  ( dsf->apfScapsSizePixels < 2 )
	{ /* LDEB(dsf->apfScapsSizePixels); */ dsf->apfScapsSizePixels= 2; }

    *pApf= dsf;
    *pFresh= 1;
    return screenFont;
    }

void appDrawInitEncodedScreenFont(	EncodedScreenFont *	esf )
    {
    esf->esfFontStruct= (APP_FONT *)0;
    esf->esfFontName[0]= '\0';
    esf->esfIsTwoByte= 0;
    esf->esfFontBorrowed= 0;
    esf->esfSymbolToByteMapping= (IndexMapping *)0;

    return;
    }

int appDrawSystemFont(	AppDrawingData *	add,
			APP_FONT *		xfs,
			int			borrowed )
    {
    PagedList *		apfl= &(add->addScreenFontList);
    DrawScreenFont *	dsf= (DrawScreenFont *)0;
    int			screenFont;

    dsf= utilPagedListClaimItemAtEnd( &screenFont, apfl );
    if  ( ! dsf )
	{ XDEB(dsf); return -1;	}

    dsf->apfEncodedFonts.esfFonts= malloc( sizeof(EncodedScreenFont) );
    if  ( ! dsf->apfEncodedFonts.esfFonts )
	{ XDEB(dsf->apfEncodedFonts.esfFonts); return -1;	}
    appDrawInitEncodedScreenFont( dsf->apfEncodedFonts.esfFonts );
    dsf->apfEncodedFonts.esfFontCount= 1;

    dsf->apfEncodedFonts.esfFonts[0].esfFontStruct= xfs;
    dsf->apfEncodedFonts.esfFonts[0].esfFontBorrowed= borrowed;

#   ifdef USE_MOTIF
    dsf->apfDisplay= add->addDisplay;
#   endif

#   ifdef USE_X11_FONTS
    addDrawX11GetFontProperties( dsf->apfEncodedFonts.esfFonts, dsf );
    strcpy( dsf->apfFontName, dsf->apfEncodedFonts.esfFonts->esfFontName );
#   endif

    return screenFont;
    }

static int appFontOpenScreenFont( AppDrawingData *		add,
				DrawScreenFont *		dsf )
    {
    int				openFullFont= 1;
    int				openScapsFont= 1;

    const int			baselinePixels= 0;

    if  ( dsf->apfPsFaceNumber < 0 || ! dsf->apfPsFontInfo )
	{ LXDEB(dsf->apfPsFaceNumber,dsf->apfPsFontInfo); return -1;	}

    /********************************************************************/
    /*  Calculate geometry shortcuts.					*/
    /********************************************************************/
    psStrikethroughGeometry( &(dsf->apfStrikethroughPositionPixels),
		    &(dsf->apfStrikethroughThicknessPixels),
		    baselinePixels, dsf->apfFullSizePixels,
		    dsf->apfPsFontInfo );

    psUnderlineGeometry( &(dsf->apfUnderlinePositionPixels),
		    &(dsf->apfUnderlineThicknessPixels),
		    baselinePixels, dsf->apfFullSizePixels,
		    dsf->apfPsFontInfo );

    psGetSuperBaseline( &(dsf->apfSuperBaseline),
		    baselinePixels, dsf->apfFullSizePixels,
		    dsf->apfPsFontInfo );
    psGetSubBaseline( &(dsf->apfSubBaseline),
		    baselinePixels, dsf->apfFullSizePixels,
		    dsf->apfPsFontInfo );

    /********************************************************************/
    /*  Open freetype fonts (if configured)				*/
    /********************************************************************/
#   ifdef USE_XFT
    if  ( ! add->addAvoidFontconfig )
	{
	if  ( openFullFont )
	    {
	    dsf->apfXftFont= appFontOpenXftFont( add, dsf,
						    dsf->apfFullSizePixels );
	    if  ( dsf->apfXftFont )
		{ openFullFont= 0; }
	    }

	if  ( openScapsFont )
	    {
	    dsf->apfXftScapsFont= appFontOpenXftFont( add, dsf,
						    dsf->apfScapsSizePixels );
	    if  ( dsf->apfXftScapsFont )
		{ openScapsFont= 0; }
	    }
	}
#   endif

#   ifdef USE_X11_FONTS
    /********************************************************************/
    /*  Open X11 fonts (if freetype is not configured, or it failed)	*/
    /********************************************************************/
    if  ( openFullFont )
	{
	if  ( ! appFontOpenX11Fonts( &(dsf->apfEncodedFonts),
					dsf, add, dsf->apfFullSizePixels ) )
	    { openFullFont= 0; }
	}

    if  ( openScapsFont )
	{
	if  ( appFontOpenX11Fonts( &(dsf->apfScapsEncodedFonts),
					dsf, add, dsf->apfScapsSizePixels ) )
	    { LDEB(1);	}

	if  ( ! openFullFont )
	    { openScapsFont= 0;	}
	}
#   endif

    if  ( openFullFont || openScapsFont )
	{
	SLLDEB(dsf->apfPsFontInfo->afiFontName,openFullFont,openScapsFont);
	return -1;
	}

    return 0;
    }

int appOpenScreenFont(		AppDrawingData *		add,
				DocumentFontList *		dfl,
				const TextAttribute *		ta )
    {
    DrawScreenFont *		dsf;
    int				fresh;
    int				screenFont;

    screenFont= appGetScreenFont( &dsf, &fresh, add, dfl, ta );
    if  ( screenFont < 0 )
	{ LDEB(screenFont); return -1;	}

    if  ( fresh )
	{
	if  ( add->addForScreenDrawing && appFontOpenScreenFont( add, dsf ) )
	    { LDEB(1); return -1;	}
	}

    return screenFont;
    }

int appFontImplementationName(	char *			target,
				int			maxlen,
				AppDrawingData *	add,
				int			screenFont )
    {
    const PagedList *		apfl= &(add->addScreenFontList);
    const AfmFontInfo *		afi;
    const DrawScreenFont *	dsf;

    dsf= (DrawScreenFont *)utilPagedListGetItemByNumber( apfl, screenFont );
    if  ( ! dsf )
	{ LXDEB(screenFont,dsf); return -1;	}
    afi= dsf->apfPsFontInfo;

#   ifdef USE_XFT
    if  ( dsf->apfXftFont )
	{
	if  ( ! afi->afiFontFileName 			||
	      strlen( afi->afiFontFileName ) >= maxlen	)
	    { return -1;	}

	strcpy( target, afi->afiFontFileName );
	return 0;
	}
#   endif

#   ifdef	USE_X11_FONTS
    if  ( dsf->apfEncodedFonts.esfFontCount > 0 )
	{
	if  ( maxlen < sizeof(dsf->apfFontName) )
	    { return -1;	}

	strcpy( target, dsf->apfFontName );

	return 0;
	}
#   endif

    return -1;
    }

