#   ifndef	APP_SCREEN_FONT_H
#   define	APP_SCREEN_FONT_H

#   include	<appGuiBase.h>

#   ifdef USE_FONTCONFIG
#	include	<ft2build.h>
#	include	FT_FREETYPE_H

#	include	<X11/Xft/Xft.h>
#	include	<X11/extensions/Xrender.h>
#   endif

#   ifdef USE_GTK
    typedef		GdkFont			(APP_FONT);
#   endif

#   ifdef USE_MOTIF
    typedef		XFontStruct		(APP_FONT);
#   endif

/************************************************************************/
/*									*/
/*  For the administration of the relation between PostScript fonts and	*/
/*  X11 fonts.								*/
/*									*/
/************************************************************************/

# define X_FONTNAME_MAX	100

typedef struct EncodedScreenFont
    {
    APP_FONT *			esfFontStruct;
    char			esfFontName[X_FONTNAME_MAX+ 1];
    unsigned char		esfIsTwoByte;
    unsigned char		esfFontBorrowed;
    const IndexMapping *	esfSymbolToByteMapping;
    } EncodedScreenFont;

typedef struct EncodedScreenFontList
    {
    EncodedScreenFont *	esfFonts;
    int			esfFontCount;
    IndexMapping	esflSymbolToFont;
    } EncodedScreenFontList;

typedef struct DrawScreenFont
    {
				/*  KEY  */
    short int			apfPsFaceNumber;
    short int			apfFullSizePixels;

				/*  DEP  */
    const AfmFontInfo *		apfPsFontInfo;
    int				apfUnderlinePositionPixels;
    int				apfUnderlineThicknessPixels;
    int				apfStrikethroughPositionPixels;
    int				apfStrikethroughThicknessPixels;
    int				apfSuperBaseline;
    int				apfSubBaseline;
    int				apfScapsSizePixels;
    int				apfXHeightPixels;
    EncodedScreenFontList	apfEncodedFonts;
    EncodedScreenFontList	apfScapsEncodedFonts;
    char			apfFontName[X_FONTNAME_MAX+ 1];

    const IndexSet *		apfUnicodesUsed;

#   if defined(USE_XFT) || defined(USE_MOTIF)
    Display *			apfDisplay;
#   endif

#   ifdef USE_XFT
    XftFont *			apfXftFont;
    XftFont *			apfXftScapsFont;
#   endif
    } DrawScreenFont;

/************************************************************************/
/*									*/
/*  Procedure declarations.						*/
/*									*/
/************************************************************************/

extern void appDrawCleanScreenFontList(	AppDrawingData *	add );
extern void appDrawStartScreenFontList(	AppDrawingData *	add );

extern int appFontOpenX11Fonts(	EncodedScreenFontList *		esfl,
				DrawScreenFont *		dsf,
				AppDrawingData *		add,
				int				sizePixels );

extern int appDrawSystemFont(		AppDrawingData *	add,
					APP_FONT *		xfs,
					int			borrowed );

extern void appDrawCleanEncodedScreenFont(	DrawScreenFont *	dsf,
						EncodedScreenFont *	esf );

extern void appDrawInitEncodedScreenFont(	EncodedScreenFont *	esf );

#   ifdef USE_X11_FONTS

extern void addDrawX11GetFontProperties(	EncodedScreenFont *	esf,
						DrawScreenFont *	dsf );
#   endif

#   ifdef USE_XFT

extern int appDrawDrawStringXft(	AppDrawingData *	add,
					const DrawScreenFont *	dsf,
					XftFont *		font,
					int			x,
					int			y,
					const char *		s,
					int			len );

extern int appDrawTextExtentsXft(	DocumentRectangle *	drText,
					int			x,
					int			y,
					AppDrawingData *	add,
					const DrawScreenFont *	dsf,
					XftFont *		hfont,
					XftFont *		vfont,
					const char *		s,
					int			len );

extern XftFont * appFontOpenXftFont(	AppDrawingData *	add,
					DrawScreenFont *	dsf,
					int			sizePixels );

#   endif
#   endif
