/************************************************************************/
/*									*/
/*  A spelling tool.							*/
/*									*/
/************************************************************************/

#   include	"appFrameConfig.h"

#   include	<stdlib.h>
#   include	<stdio.h>
#   include	<stddef.h>
#   include	<string.h>
#   include	<appDebugon.h>

#   include	<errno.h>

#   include	"appSpellTool.h"
#   include	<appSystem.h>
#   include	<appFrame.h>

#   include	<charnames.h>

#   ifdef USE_MOTIF
#	include	<X11/keysym.h>
#   endif

/************************************************************************/
/*									*/
/*  Resources for the spell tool.					*/
/*									*/
/************************************************************************/

typedef struct LocaleName
    {
    const char *	lnLocale;
    const char *	lnLocaleName;
    } LocaleName;

/* From ../tedPackage/languages.txt */
static LocaleName AppSpellToolLocaleNames[]=
{
    { "en_US", "American English",		},

    { "nl_NL", "Dutch",				},
    { "en_GB", "British English",		},
    { "de_DE", "German",			},
    { "es_ES", "Spanish",			},
    { "es_AR", "Argentinian Spanish",		},
    { "pt_PT", "Portuguese",			},
    { "pt_BR", "Brazilian Portuguese",		},
    { "fr_FR", "French",			},
    { "it_IT", "Italian",			},
    { "cs_CZ", "Czech",				},
    { "da_DK", "Danish",			},
    { "sv_SE", "Swedish",			},
    { "no_NO", "Norwegian",			},
    { "pl_PL", "Polish",			},
    { "ru_RU", "Russian",			},
    { "sk_SK", "Slovak",			},
    { "sl_SI", "Slovene",			},
    { "hu_HU", "Hungarian",			},
    { "mg_MG", "Malagasi",			},
    { "oc_FR", "Occitan",			},
    { "eu_FR", "Basque",			},
    { "bg_BG", "Bulgarian",			},
    { "hr_HR", "Croatian",			},
    { "et_EE", "Estonian",			},
    { "fi_FI", "Finnish",			},
};

typedef struct AppSpellToolResources
    {
    char *		astrDictionary;
    char *		astrLearn;
    char *		astrForget;
    char *		astrGuesses;
    char *		astrIgnore;
    char *		astrFindNext;
    char *		astrGuess;
    char *		astrCorrect;
    char *		astrLower;
    char *		astrClose;
    char *		astrPrivateDictionaries;
    char *		astrSystemDictionaries;
    char *		astrNoDicts;

    char *		astrErrors[SCerr_COUNT];

    LocaleName *	astrLocaleNames;
    int			astrLocaleNameCount;
    } AppSpellToolResources;

static AppConfigurableResource APP_SpellToolResourceTable[]=
    {
    /************************/
    /*  Spell Tool.		*/
    /************************/
    APP_RESOURCE( "spellToolDictTitle",
		offsetof(AppSpellToolResources,astrDictionary),
		"Dictionary" ),
    APP_RESOURCE( "spellToolLearn",
		offsetof(AppSpellToolResources,astrLearn),
		"Learn" ),
    APP_RESOURCE( "spellToolForget",
		offsetof(AppSpellToolResources,astrForget),
		"Forget" ),
    APP_RESOURCE( "spellToolGuesses",
		offsetof(AppSpellToolResources,astrGuesses),
		"Guesses" ),
    APP_RESOURCE( "spellToolIgnore",
		offsetof(AppSpellToolResources,astrIgnore),
		"Ignore" ),
    APP_RESOURCE( "spellToolFindNext",
		offsetof(AppSpellToolResources,astrFindNext),
		"Find Next" ),
    APP_RESOURCE( "spellToolGuess",
		offsetof(AppSpellToolResources,astrGuess),
		"Guess" ),
    APP_RESOURCE( "spellToolCorrect",
		offsetof(AppSpellToolResources,astrCorrect),
		"Correct" ),
    APP_RESOURCE( "spellToolNoDicts",
		offsetof(AppSpellToolResources,astrNoDicts),
		"None" ),

    APP_RESOURCE( "spellToolLower",
		offsetof(AppSpellToolResources,astrLower),
		"Lower" ),
    APP_RESOURCE( "spellToolClose",
		offsetof(AppSpellToolResources,astrClose),
		"Close" ),

#   ifdef __VMS
    APP_RESOURCE( "spellToolPrivateDicts",
		offsetof(AppSpellToolResources,astrPrivateDictionaries),
		"_Dictionaries" ),
#   else
    APP_RESOURCE( "spellToolPrivateDicts",
		offsetof(AppSpellToolResources,astrPrivateDictionaries),
		".Dictionaries" ),
#   endif
    APP_RESOURCE( "spellToolSystemDicts",
		offsetof(AppSpellToolResources,astrSystemDictionaries),
		INDDIR ),

    APP_RESOURCE( "spellToolDirNoAccess",
	    offsetof(AppSpellToolResources,astrErrors[SCerrDIR_NO_ACCESS]),
	    "Could not access directory for dictionaries." ),
    APP_RESOURCE( "spellToolDirNoSuchDir",
	    offsetof(AppSpellToolResources,astrErrors[SCerrDIR_DOES_NOT_EXIST]),
	    "This directory could not be found." ),
    APP_RESOURCE( "spellToolDirNoDicts",
	    offsetof(AppSpellToolResources,astrErrors[SCerrDIR_HAS_NO_DICTS]),
	    "No dictionaries were found in this directory." ),
    APP_RESOURCE( "spellToolSysDictNoAccess",
	    offsetof(AppSpellToolResources,astrErrors[SCerrDICT_NO_ACCESS]),
	    "Could not read system dictionary." ),
    APP_RESOURCE( "spellToolPrivDictDirNotMade",
	    offsetof(AppSpellToolResources,astrErrors[SCerrPRIVATE_DIR_NOT_MADE]),
	    "Could not make private directory." ),
    APP_RESOURCE( "spellToolPrivDictNoAccess",
	    offsetof(AppSpellToolResources,astrErrors[SCerrPRIVATE_DICT_NO_ACCESS]),
	    "Could not read private dictionary." ),
    APP_RESOURCE( "spellToolPrivDictWrongFormat",
	    offsetof(AppSpellToolResources,astrErrors[SCerrPRIVATE_DICT_WRONG_FORMAT]),
	    "Private dictionary has an illegal format." ),
    };

/************************************************************************/
/*									*/
/*  Represents a spell tool.						*/
/*									*/
/************************************************************************/
#   define	FILEL	400

typedef struct AppSpellTool
    {
    EditApplication *		astApplication;

    AppSpellToolResources *	astResources;

    APP_WIDGET			astTopWidget;
    APP_WIDGET			astMainWidget;

    AppOptionmenu		astDictionaryOptionmenu;
    APP_WIDGET			astGuessList;
    APP_WIDGET			astWordText;

    APP_WIDGET			astLearnButton;
    APP_WIDGET			astForgetButton;
    APP_WIDGET			astCorrectButton;
    APP_WIDGET			astGuessButton;
    APP_WIDGET			astIgnoreButton;
    APP_WIDGET			astLowerButton;
    APP_WIDGET			astCloseButton;

    AppToolDestroy		astDestroy;
    SpellChecker *		astSpellChecker;
    int				astCurrentDictionary;
    void *			astTarget;
    } AppSpellTool;

/************************************************************************/
/*									*/
/*  Adapt buttons to different situations..				*/
/*									*/
/************************************************************************/

static void appSpellToolSomethingFound(	AppSpellTool *	ast,
					int		yes_no )
    {
    appGuiEnableWidget( ast->astLearnButton, yes_no );
    appGuiEnableWidget( ast->astIgnoreButton, yes_no );

    return;
    }

static void appSpellToolGotAlternative(	AppSpellTool *	ast,
					int		yes_no )
    {
    appGuiEnableWidget( ast->astForgetButton, yes_no );
    appGuiEnableWidget( ast->astCorrectButton, yes_no );
    appGuiEnableWidget( ast->astGuessButton, yes_no );

    return;
    }

/************************************************************************/
/*									*/
/*  Complain								*/
/*									*/
/************************************************************************/

static void appSpellToolComplain(	int			error,
					const char *		subject,
					void *			voidast )
    {
    AppSpellTool *		ast= (AppSpellTool *)voidast;
    const char *		message= "??";
    AppSpellToolResources *	astr= ast->astResources;

    if  ( error > 0 && error < SCerr_COUNT )
	{ message= astr->astrErrors[error]; }

    if  ( subject )
	{
	appQuestionRunErrorDialog( ast->astApplication,
		    ast->astTopWidget, (APP_WIDGET)0, message );
	}
    else{
	appQuestionRunSubjectErrorDialog( ast->astApplication,
		    ast->astTopWidget, (APP_WIDGET)0, subject, message );
	}

    return;
    }

/************************************************************************/
/*									*/
/*  Look for the next unknown word.					*/
/*									*/
/************************************************************************/

static void appSpellToolFindNext(	AppSpellTool *	ast )
    {
    SpellChecker *		sc;
    SpellDictionary *		sd;

    if  ( ast->astCurrentDictionary < 0 )
	{ LDEB(ast->astCurrentDictionary); return;	}
    if  ( ! ( sc= ast->astSpellChecker ) )
	{ XDEB(ast->astSpellChecker); return;	}

    sd= sc->scDictionaries+ ast->astCurrentDictionary;

    appStringToTextWidget( ast->astWordText, "" );
    appGuiEmptyListWidget( ast->astGuessList );

    if  ( (*sc->scOpenDict)( sd, appSpellToolComplain, (void *)ast, sc ) )
	{ return;	}

    if  ( ! (*sc->scFindNext)( ast->astTarget, sc, sd ) )
	{ appSpellToolSomethingFound( ast, 1 );	}
    else{ appSpellToolSomethingFound( ast, 0 );	}

    appSpellToolGotAlternative( ast, 0 );
    }

/************************************************************************/
/*									*/
/*  'Learn/Forget/Ignore' button has been pushed.			*/
/*									*/
/************************************************************************/

static APP_BUTTON_CALLBACK_H( appSpellToolLearnPushed, w, voidast )
    {
    AppSpellTool *		ast= (AppSpellTool *)voidast;
    SpellChecker *		sc;
    SpellDictionary *		sd;
    char *			word;

    if  ( ast->astCurrentDictionary < 0 )
	{ LDEB(ast->astCurrentDictionary); return;	}
    if  ( ! ( sc= ast->astSpellChecker ) )
	{ XDEB(ast->astSpellChecker); return;	}
    if  ( ! sc->scLearnWord )
	{ XDEB(sc->scLearnWord); return;		}

    sd= sc->scDictionaries+ ast->astCurrentDictionary;

    word= appGetStringFromTextWidget( ast->astWordText );

    (*sc->scLearnWord)( sd, appSpellToolComplain, (void *)ast, sc, word );

    appFreeStringFromTextWidget( word );

    appSpellToolFindNext( ast );

    return;
    }

static APP_BUTTON_CALLBACK_H( appSpellToolForgetPushed, w, voidast )
    {
    AppSpellTool *		ast= (AppSpellTool *)voidast;
    SpellChecker *		sc;
    SpellDictionary *		sd;
    char *			word;

    if  ( ast->astCurrentDictionary < 0 )
	{ LDEB(ast->astCurrentDictionary); return;	}
    if  ( ! ( sc= ast->astSpellChecker ) )
	{ XDEB(ast->astSpellChecker); return;	}
    if  ( ! sc->scForgetWord )
	{ XDEB(sc->scForgetWord); return;		}

    sd= sc->scDictionaries+ ast->astCurrentDictionary;

    word= appGetStringFromTextWidget( ast->astWordText );

    (*sc->scForgetWord)( sd, appSpellToolComplain, (void *)ast, sc, word );

    appFreeStringFromTextWidget( word );

    appGuiEnableWidget( ast->astForgetButton, 0 );
    appGuiEnableWidget( ast->astLearnButton, 1 );

    return;
    }

static APP_BUTTON_CALLBACK_H( appSpellToolIgnorePushed, w, voidast )
    {
    AppSpellTool *		ast= (AppSpellTool *)voidast;
    SpellChecker *		sc;
    SpellDictionary *		sd;
    char *			word;

    if  ( ast->astCurrentDictionary < 0 )
	{ LDEB(ast->astCurrentDictionary); return;	}
    if  ( ! ( sc= ast->astSpellChecker ) )
	{ XDEB(ast->astSpellChecker); return;	}
    if  ( ! sc->scIgnoreWord )
	{ XDEB(sc->scIgnoreWord); return;		}

    sd= sc->scDictionaries+ ast->astCurrentDictionary;

    word= appGetStringFromTextWidget( ast->astWordText );

    (*sc->scIgnoreWord)( sd, appSpellToolComplain, (void *)ast, sc, word );

    appFreeStringFromTextWidget( word );

    appSpellToolFindNext( ast );

    return;
    }

/************************************************************************/
/*									*/
/*  'Find Next' button has been pushed.					*/
/*									*/
/************************************************************************/

static APP_BUTTON_CALLBACK_H( appSpellToolFindNextPushed, w, voidast )
    {
    AppSpellTool *		ast= (AppSpellTool *)voidast;

    appSpellToolFindNext( ast );

    return;
    }

/************************************************************************/
/*									*/
/*  The user typed something in the Coreection, Turn on the the		*/
/*  'Correct' button.							*/
/*									*/
/************************************************************************/

static APP_TXTYPING_CALLBACK_H( appSpellCorrectionTyped, w, voidast )
    {
    AppSpellTool *	ast= (AppSpellTool *)voidast;

    appSpellToolGotAlternative( ast, 1 );

    return;
    }

/************************************************************************/
/*									*/
/*  'Correct' button has been pushed.					*/
/*  Or a double click on the listbox with guesses.			*/
/*									*/
/************************************************************************/

static APP_BUTTON_CALLBACK_H( appSpellToolCorrect, w, voidast )
    {
    AppSpellTool *	ast= (AppSpellTool *)voidast;
    SpellChecker *	sc;
    char *		guess;

    if  ( ! ( sc= ast->astSpellChecker ) )
	{ XDEB(ast->astSpellChecker); return;	}
    if  ( ! sc->scCorrect )
	{ XDEB(sc->scCorrect); return;	}

    guess= appGetStringFromTextWidget( ast->astWordText );

    (*sc->scCorrect)( ast->astTarget, guess );

    appFreeStringFromTextWidget( guess );

    appSpellToolFindNext( ast );

    return;
    }

/************************************************************************/
/*									*/
/*  The 'Guess' button has been pushed.					*/
/*									*/
/************************************************************************/

static APP_BUTTON_CALLBACK_H( appSpellGuessButtonPushed, w, voidast )
    {
    AppSpellTool *	ast= (AppSpellTool *)voidast;
    char *		word;

    word= appGetStringFromTextWidget( ast->astWordText );

    appSpellMakeGuesses( voidast, word );

    appFreeStringFromTextWidget( word );

    return;
    }

/************************************************************************/
/*  A guess in the list has been selected.				*/
/************************************************************************/

static APP_LIST_CALLBACK_H( appSpellGuessChosen, w, voidast, voidlcs )
    {
    AppSpellTool *		ast= (AppSpellTool *)voidast;

    char *			text;

    text= appGuiGetStringFromListCallback( w, voidlcs );
    if  ( text )
	{
	appStringToTextWidget( ast->astWordText, text );
	appFreeStringFromListCallback( text );
	}

    appSpellToolGotAlternative( ast, 1 );

    return;
    }

/************************************************************************/
/*									*/
/*  Make the dictionary part of the spelling tool.			*/
/*									*/
/************************************************************************/

static APP_WIDGET appSpellMakeDictionaryFrame( APP_WIDGET	parent,
					AppSpellToolResources * astr,
					AppSpellTool *		ast )
    {
    APP_WIDGET		frame;
    APP_WIDGET		paned;

    APP_WIDGET		buttonRow;

    appMakeColumnFrameInColumn( &frame, &paned,
					    parent, astr->astrDictionary );

    appMakeOptionmenuInColumn( &(ast->astDictionaryOptionmenu), paned );

    appInspectorMakeButtonRow( &buttonRow, paned,
		&(ast->astLearnButton), &(ast->astForgetButton),
		astr->astrLearn, astr->astrForget,
		appSpellToolLearnPushed, appSpellToolForgetPushed,
		(void *)ast );

    return frame;
    }

/************************************************************************/
/*									*/
/*  A spell tool must be destroyed.					*/
/*									*/
/************************************************************************/

static void appDestroySpellTool(	AppSpellTool *	ast )
    {
    if  ( ast->astDestroy )
	{ (*ast->astDestroy)( ast->astTarget );	}

    appDestroyShellWidget( ast->astTopWidget );

    free( ast );

    return;
    }

static APP_CLOSE_CALLBACK_H( appCloseSpellTool, w, voidast )
    {
    AppSpellTool *	ast= (AppSpellTool *)voidast;

    appDestroySpellTool( ast );

    return;
    }

static APP_BUTTON_CALLBACK_H( appLowerSpellTool, w, voidast )
    {
    AppSpellTool *	ast= (AppSpellTool *)voidast;
    APP_WIDGET		shell= ast->astTopWidget;

    appGuiLowerShellWidget( shell );

    return;
    }

static APP_BUTTON_CALLBACK_H( appSpellToolCloseButtonPushed, w, voidast )
    {
    AppSpellTool *	ast= (AppSpellTool *)voidast;

    appDestroySpellTool( ast );

    return;
    }

#   ifdef USE_MOTIF
static void appSpellToolCancel(		APP_WIDGET		w,
					void *			voidast,
					XEvent *		event,
					Boolean *		pRefused )
    {
    if  ( event->type == KeyRelease )
	{
	XKeyEvent *		kevent= &(event->xkey);
	AppSpellTool *		ast= (AppSpellTool *)voidast;

	if  ( XKeysymToKeycode( XtDisplay(w), XK_Escape ) == kevent->keycode )
	    { appDestroySpellTool( ast ); *pRefused= 0; return; }
	}

    *pRefused= 1; return;
    }
#   endif

/************************************************************************/
/*									*/
/*  A dictionary has been selected.					*/
/*									*/
/************************************************************************/

static APP_OITEM_CALLBACK_H( appSpellDictionaryChosen, w, voidast )
    {
    AppSpellTool *	ast= (AppSpellTool *)voidast;
    SpellChecker *	sc;

    int			dict;

    if  ( ! ( sc= ast->astSpellChecker ) )
	{ XDEB(ast->astSpellChecker); return;	}

    dict= appGuiGetOptionmenuItemIndex( &(ast->astDictionaryOptionmenu), w );
    if  ( dict < 0 || dict >= sc->scDictionaryCount )
	{ LLDEB(dict,sc->scDictionaryCount); return;	}

    ast->astCurrentDictionary= dict;

    return;
    }

/************************************************************************/
/*									*/
/*  Fill the list of dictionaries.					*/
/*									*/
/************************************************************************/

static void appSpellFillDictionaryMenu(		AppSpellTool *		ast,
						AppSpellToolResources *	astr )
    {
    SpellChecker *	sc;
    SpellDictionary *	sd;

    int			dict;

    if  ( ! ( sc= ast->astSpellChecker ) )
	{ XDEB(ast->astSpellChecker); return;	}

    appEmptyOptionmenu( &(ast->astDictionaryOptionmenu) );

    sd= sc->scDictionaries;
    for ( dict= 0; dict < sc->scDictionaryCount; sd++, dict++ )
	{
	appAddItemToOptionmenu(
		    &(ast->astDictionaryOptionmenu), sd->sdLocaleLabel,
		    appSpellDictionaryChosen, (void *)ast );
	}

    if  ( sc->scDictionaryCount == 0 )
	{
	(void) appAddItemToOptionmenu( &(ast->astDictionaryOptionmenu),
			    astr->astrNoDicts,
			    appSpellDictionaryChosen, (void *)ast );

	appGuiEnableWidget( ast->astTopWidget, 0 );
	}

    appSetOptionmenu( &(ast->astDictionaryOptionmenu), 0 );
    ast->astCurrentDictionary= 0;

    appOptionmenuRefreshWidth( &(ast->astDictionaryOptionmenu) );
    }

/************************************************************************/
/*									*/
/*  Make the listbox with guesses.					*/
/*									*/
/************************************************************************/

static APP_WIDGET appSpellGuessList(	APP_WIDGET		parent,
					AppSpellToolResources * astr,
					AppSpellTool *		ast )
    {
    APP_WIDGET		label;
    APP_WIDGET		list;

    const int		visibleItems= 6;

    appMakeLabelInColumn( &label, parent, astr->astrGuesses );

    appGuiMakeListInColumn( &list, parent, visibleItems,
		    appSpellGuessChosen, appSpellToolCorrect, (void *)ast );

    return list;
    }

/************************************************************************/
/*									*/
/*  Make the button part of the spelling Tool.				*/
/*									*/
/************************************************************************/

static void appSpellMakeButtonRows(	APP_WIDGET		spellForm,
					AppSpellToolResources * astr,
					AppSpellTool *		ast )
    {
    APP_WIDGET		buttonRow;

    APP_WIDGET		findNextButton;

    appInspectorMakeButtonRow( &buttonRow, spellForm,
		&(ast->astIgnoreButton), &(findNextButton),
		astr->astrIgnore, astr->astrFindNext,
		appSpellToolIgnorePushed, appSpellToolFindNextPushed,
		(void *)ast );

    appInspectorMakeButtonRow( &buttonRow, spellForm,
		&(ast->astGuessButton), &(ast->astCorrectButton),
		astr->astrGuess, astr->astrCorrect,
		appSpellGuessButtonPushed, appSpellToolCorrect,
		(void *)ast );

    appInspectorMakeButtonRow( &buttonRow, spellForm,
		&(ast->astLowerButton), &(ast->astCloseButton),
		astr->astrLower, astr->astrClose,
		appLowerSpellTool, appSpellToolCloseButtonPushed,
		(void *)ast );

    return;
    }

/************************************************************************/
/*									*/
/*  Resolve locale names.						*/
/*									*/
/************************************************************************/

static int appSpellGetLocaleNames(		EditApplication *	ea,
						SpellChecker *		sc )
    {
    static int				gotResources;

    static LocaleName *			localeNames= (LocaleName *)0;
    static int				localeNameCount;
    static AppConfigurableResource *	localeRes= (AppConfigurableResource *)0;

    int					i;

    if  ( ! gotResources )
	{
	localeNames= malloc( sc->scDictionaryCount* sizeof(LocaleName) );
	if  ( ! localeNames )
	    { LXDEB(sc->scDictionaryCount,localeNames); return -1; }
	localeRes= malloc( sc->scDictionaryCount*
					    sizeof(AppConfigurableResource) );
	if  ( ! localeRes )
	    { LXDEB(sc->scDictionaryCount,localeRes); return -1; }

	for ( i= 0; i < sc->scDictionaryCount; i++ )
	    {
	    int			j;
	    char *		name;
	    const char *	def;

	    localeNames[i].lnLocale= strdup( sc->scDictionaries[i].sdLocale );
	    if  ( ! localeNames[i].lnLocale )
		{ XDEB(localeNames[i].lnLocale); return -1;	}

	    name= malloc( 5+ strlen(sc->scDictionaries[i].sdLocale)+ 1 );
	    if  ( ! name )
		{ XDEB(name); return -1;	}

	    sprintf( name, "NAME_%s", sc->scDictionaries[i].sdLocale );
	    def=  sc->scDictionaries[i].sdLocale;

	    for ( j= 0;
		  j < sizeof(AppSpellToolLocaleNames)/sizeof(LocaleName);
		  j++ )
		{
		if  ( ! strcmp( AppSpellToolLocaleNames[j].lnLocale,
					    sc->scDictionaries[i].sdLocale ) )
		    {
		    def= AppSpellToolLocaleNames[j].lnLocaleName;
		    break;
		    }
		}

	    APP_SET_RESOURCE( localeRes+ i, name,
		    i* sizeof(LocaleName)+ offsetof(LocaleName,lnLocaleName),
		    def );
	    }

	appGuiGetResourceValues( &gotResources, ea, (void *)localeNames,
					    localeRes, sc->scDictionaryCount );
	localeNameCount= sc->scDictionaryCount;
	gotResources= 1;
	}

    for ( i= 0; i < sc->scDictionaryCount; i++ )
	{
	int		j;

	for ( j= 0; j < localeNameCount; j++ )
	    {
	    int		n= ( i+ j ) % localeNameCount;

	    if  ( ! strcmp( localeNames[n].lnLocale,
					sc->scDictionaries[i].sdLocale ) )
		{
		indSpellDictionarySetLocaleLabel( &(sc->scDictionaries[i]),
						localeNames[n].lnLocaleName );
		break;
		}
	    }
	}

    return 0;
    }

/************************************************************************/
/*									*/
/*  make a spell tool.							*/
/*									*/
/************************************************************************/

void * appMakeSpellTool(	APP_WIDGET		spellOption,
				EditApplication *	ea,
				const char *		widgetName,
				APP_BITMAP_IMAGE	iconPixmap,
				APP_BITMAP_MASK		iconMask,
				AppToolDestroy		destroy,
				SpellChecker *		sc,
				void *			target )
    {
    AppSpellTool *			ast;
    APP_WIDGET				dictionaryFrame;
    const int				userResizable= 1;

    static AppSpellToolResources	astr;
    static int				gotResources;

    if  ( ! gotResources )
	{
	appGuiGetResourceValues( &gotResources, ea, (void *)&astr,
					APP_SpellToolResourceTable,
					sizeof(APP_SpellToolResourceTable)/
					sizeof(AppConfigurableResource) );
	}

    ast= (AppSpellTool *)malloc( sizeof(AppSpellTool) );
    if  ( ! ast )
	{ XDEB(ast); return (void *)0;	}

    ast->astApplication= ea;
    ast->astResources= &astr;
    ast->astSpellChecker= sc;

    ast->astDestroy= destroy;
    ast->astTarget= target;
    sc->scPrivateDirectory= astr.astrPrivateDictionaries;
    sc->scSystemDirectory= astr.astrSystemDictionaries;

    ast->astCurrentDictionary= -1;

    appInitOptionmenu( &(ast->astDictionaryOptionmenu) );

    (*sc->scSetup)( sc, appSpellToolComplain, ast );

    if  ( appSpellGetLocaleNames( ea, sc ) )
	{ LDEB(sc->scDictionaryCount);	}

    appMakeVerticalTool( &(ast->astTopWidget), &(ast->astMainWidget), ea,
	    iconPixmap, iconMask,
	    widgetName, userResizable, spellOption,
	    appCloseSpellTool, (void *)ast );

    dictionaryFrame= appSpellMakeDictionaryFrame( ast->astMainWidget,
								&astr, ast );
    ast->astGuessList= appSpellGuessList( ast->astMainWidget, &astr, ast );

    appMakeTextInColumn( &(ast->astWordText), ast->astMainWidget, 0, 1 );

    appGuiSetTypingCallbackForText( ast->astWordText,
				    appSpellCorrectionTyped, (void *)ast );

    appSpellMakeButtonRows( ast->astMainWidget, &astr, ast );

#   ifdef USE_MOTIF
    XtInsertEventHandler( ast->astGuessList, KeyReleaseMask, False,
				appSpellToolCancel, (void *)ast, XtListHead );
    XtInsertEventHandler( ast->astWordText, KeyReleaseMask, False,
				appSpellToolCancel, (void *)ast, XtListHead );
    XtInsertEventHandler( ast->astLearnButton, KeyReleaseMask, False,
				appSpellToolCancel, (void *)ast, XtListHead );
    XtInsertEventHandler( ast->astForgetButton, KeyReleaseMask, False,
				appSpellToolCancel, (void *)ast, XtListHead );
    XtInsertEventHandler( ast->astCorrectButton, KeyReleaseMask, False,
				appSpellToolCancel, (void *)ast, XtListHead );
    XtInsertEventHandler( ast->astGuessButton, KeyReleaseMask, False,
				appSpellToolCancel, (void *)ast, XtListHead );
    XtInsertEventHandler( ast->astIgnoreButton, KeyReleaseMask, False,
				appSpellToolCancel, (void *)ast, XtListHead );
#   endif

    appSpellFillDictionaryMenu( ast, &astr );

    appShowShellWidget( ea, ast->astTopWidget );

    appOptionmenuRefreshWidth( &(ast->astDictionaryOptionmenu) );

    appSpellToolSomethingFound( ast, 0 );
    appSpellToolGotAlternative( ast, 0 );

    return (void *)ast;
    }

/************************************************************************/
/*									*/
/*  Draw a spell tool to front.						*/
/*									*/
/************************************************************************/

void appShowSpellTool(		void *	voidast	)
    {
    AppSpellTool *		ast= (AppSpellTool *)voidast;

    appStringToTextWidget( ast->astWordText, "" );
    appGuiEmptyListWidget( ast->astGuessList );

    appShowShellWidget( ast->astApplication, ast->astTopWidget );

    appSpellToolSomethingFound( ast, 0 );
    appSpellToolGotAlternative( ast, 0 );

    }

void appSpellMakeGuesses(	void *		voidast,
				const char *	word )
    {
    AppSpellTool *		ast= (AppSpellTool *)voidast;
    SpellChecker *		sc;
    SpellDictionary *		sd;

    int				count;
    int				i;
    static IndGuessList		igl;

    if  ( ast->astCurrentDictionary < 0 )
	{ LDEB(ast->astCurrentDictionary); return;	}
    if  ( ! ( sc= ast->astSpellChecker ) )
	{ XDEB(ast->astSpellChecker); return;	}
    if  ( ! sc->scGuessWord )
	{ XDEB(sc->scGuessWord); return;		}

    sd= sc->scDictionaries+ ast->astCurrentDictionary;

    count= (*sc->scGuessWord)( &igl, sd,
				appSpellToolComplain, (void *)ast, sc, word );
    if  ( count < 0 )
	{ LDEB(count); return;	}

    appGuiEmptyListWidget( ast->astGuessList );
    appStringToTextWidget( ast->astWordText, (const char *)word );

    for ( i= 0; i < count; i++ )
	{
	appGuiAddValueToListWidget( ast->astGuessList, -1,
					(char *)igl.iglGuesses[i].igsWord );
	}

    appSpellToolGotAlternative( ast, 0 );
    }

void appEnableSpellTool(	void *	voidast,
				int	enabled )
    {
    AppSpellTool *	ast= (AppSpellTool *)voidast;

    appGuiEnableWidget( ast->astMainWidget, enabled != 0 );

    return;
    }
