/************************************************************************/
/*									*/
/*  Windows Meta File specific definitions.				*/
/*									*/
/************************************************************************/

#   include	<appGuiBase.h>

#   include	<bitmap.h>
#   include	<utilDocFont.h>
#   include	<utilPostscriptFace.h>
#   include	<appDraw.h>
#   include	<appImage.h>
#   include	<sioGeneral.h>

#   include	<appMetafile.h>
#   include	<appMatchFont.h>
#   include	<bmWmf.h>

typedef struct LogicalBrush
    {
    int			lbStyle;
#			define		BS_SOLID	0
#			define		BS_NULL		1
#			define		BS_HOLLOW	BS_NULL
#			define		BS_HATCHED	2
#			define		BS_PATTERN	3
#			define		BS_INDEXED	4
#			define		BS_DIBPATTERN	5
    RGB8Color		lbColor;
    int			lbHatch;
#			define HS_HORIZONTAL       0
#			define HS_VERTICAL         1
#			define HS_FDIAGONAL        2
#			define HS_BDIAGONAL        3
#			define HS_CROSS            4
#			define HS_DIAGCROSS        5

    APP_BITMAP_IMAGE	lbTilePixmap;
    } LogicalBrush;

typedef struct PatternBrush
    {
    int			pbType;		/*  == 5	*/
    int			pbUsage;	/*  == 0	*/
    AppBitmapImage	pbAbi;
    APP_BITMAP_IMAGE	pbTilePixmap;
    } PatternBrush;

typedef struct LogicalFont
    {
    int			lfHeight;
    int			lfWidth;
    int			lfEscapement;
    int			lfOrientation;
    int			lfWeight;
    unsigned char	lfItalic;
    unsigned char	lfUnderline;
    unsigned char	lfStrikeOut;
    unsigned char	lfCharSet;
    unsigned char	lfOutPrecision;
    unsigned char	lfClipPrecision;
    unsigned char	lfQuality;
    unsigned char	lfPitchAndFamily;
    char		lfFaceName[32];

    /************************************/
    /*  Implementation			*/
    /************************************/
    TextAttribute	lfTextAttribute;
    int			lfPrivateFont;
    } LogicalFont;

typedef union MetafileSpecific
    {
    LogicalBrush	mfsLogicalBrush;
    LogicalPen		mfsLogicalPen;
    LogicalFont		mfsLogicalFont;
    PatternBrush	mfsPatternBrush;
    } MetafileSpecific;

typedef struct MetaFileObject
    {
    int			mfoType;
    MetafileSpecific	mfoSpecific;
    } MetaFileObject;

/************************************************************************/
/*									*/
/*  Refer to Petzold, Charles: "Programming Windows 95", Microsoft	*/
/*  Press, Redmond, WA, 1996 pp138..					*/
/*									*/
/*  Terminology is extremely confusing: The window is the system of	*/
/*  'logical' user coordinates. The viewport is the sytem of 'hard'	*/
/*  device coordinates. Viewport=Device coordinates usually are in	*/
/*  pixels.								*/
/*									*/
/************************************************************************/

typedef struct DeviceCoordinateTransform
    {
    int					dct_xWinOrg;
    int					dct_yWinOrg;
    int					dct_xWinExt;
    int					dct_yWinExt;

    int					dct_xViewOrg;
    int					dct_yViewOrg;
    int					dct_xViewExt;
    int					dct_yViewExt;

    struct DeviceCoordinateTransform *	dctPushed;
    } DeviceCoordinateTransform;

# define DC_xWinOrg(dc)		((dc)->dcTransform.dct_xWinOrg)
# define DC_yWinOrg(dc)		((dc)->dcTransform.dct_yWinOrg)
# define DC_xWinExt(dc)		((dc)->dcTransform.dct_xWinExt)
# define DC_yWinExt(dc)		((dc)->dcTransform.dct_yWinExt)

# define DC_xViewOrg(dc)	((dc)->dcTransform.dct_xViewOrg)
# define DC_yViewOrg(dc)	((dc)->dcTransform.dct_yViewOrg)
# define DC_xViewExt(dc)	((dc)->dcTransform.dct_xViewExt)
# define DC_yViewExt(dc)	((dc)->dcTransform.dct_yViewExt)

# define DC_xViewport(x,dc) ( ( ( x-DC_xWinOrg(dc) )* DC_xViewExt(dc) ) / \
					    DC_xWinExt(dc)+ DC_xViewOrg(dc) )
# define DC_yViewport(y,dc) ( ( ( y-DC_yWinOrg(dc) )* DC_yViewExt(dc) ) / \
					    DC_yWinExt(dc)+ DC_yViewOrg(dc) )

# define DC_wViewport(w,dc) ( ( (w)* DC_xViewExt(dc) ) / DC_xWinExt(dc) )
# define DC_hViewport(h,dc) ( ( (h)* DC_yViewExt(dc) ) / DC_yWinExt(dc) )

typedef struct DeviceContext
    {
    int				dcMapMode;
    int				dcDrawBorders;
    int				dcFillInsides;
    int				dcROP2;

    int				dcClipX0;
    int				dcClipY0;
    int				dcClipX1;
    int				dcClipY1;

    int				dcFillHatched;
    int				dcFillPattern;

    DocumentFontList		dcFontList;
    const PostScriptFontList *	dcPostScriptFontList;
    AppDrawingData		dcDrawingData;

    LogicalPen			dcPen;
    LogicalBrush		dcBrush;
    LogicalFont			dcFont;

    RGB8Color			dcBkColor;
    RGB8Color			dcTextColor;
    int				dcBkMode;

    int				dcRedSet;
    int				dcGreenSet;
    int				dcBlueSet;
    TextAttribute		dcTextAttributeSet;
    LogicalPen			dcPenSet;

    int				dcExtraTextSpacing;
    int				dcJustificationAmount;
    int				dcJustificationSpaces;
    int				dcTextAlignment;
    long			dcStretchBltMode;

    int				dcX;
    int				dcY;

    char *			dcString;

    DeviceCoordinateTransform	dcTransform;
    int				dcTwipsWide;
    int				dcTwipsHigh;

    MetaFileObject *		dcObjects;
    int				dcObjectCount;

    APP_POINT *			dcPoints;

    const AfmFontInfo *		dcAfi;
    int				dcFontEncoding;
    } DeviceContext;

# define DCX_xViewport(x,dcx)	(DC_xViewport(x,&(dcx->dcxDeviceContext)))
# define DCX_yViewport(y,dcx)	(DC_yViewport(y,&(dcx->dcxDeviceContext)))
# define DCX_wViewport(w,dcx)	(DC_wViewport(w,&(dcx->dcxDeviceContext)))
# define DCX_hViewport(h,dcx)	(DC_hViewport(h,&(dcx->dcxDeviceContext)))

typedef struct DeviceContextX11
    {
    DeviceContext	dcxDeviceContext;

    int			dcxPixelsWide;
    int			dcxPixelsHigh;

    APP_COLOR_RGB	dcxPenColor;
    APP_COLOR_RGB	dcxBrushColor;

    APP_COLOR_RGB	dcxBkColor;
    APP_COLOR_RGB	dcxTextColor;
    int			dcxFillTiled;
    int			dcxFillHatched;

    int			dcxScreenFont;
    } DeviceContextX11;

/************************************************************************/
/*									*/
/*  Routine declarations.						*/
/*									*/
/************************************************************************/

extern int appMetaPlayWmfX11(	SimpleInputStream *	sis,
				void **			pPrivate,
				AppColors *		ac,
				AppDrawingData *	add,
				APP_WINDOW		pixmap,
				int			mapMode,
				int			xWinExt,
				int			yWinExt,
				int			pixelsWide,
				int			pixelsHigh,
				int			twipsWide,
				int			twipsHigh );

extern int appMetaPlayEmfX11(	SimpleInputStream *	sis,
				void **			pPrivate,
				AppColors *		ac,
				AppDrawingData *	add,
				APP_WINDOW		pixmap,
				int			mapMode,
				int			xExt,
				int			yExt,
				int			pixelsWide,
				int			pixelsHigh,
				int			twipsWide,
				int			twipsHigh );

extern int appMetaPlayFilePs(	PrintingState *			ps,
				SimpleInputStream *		sis,
				const PostScriptFontList *	psfl,
				int				useFilters,
				int				indexedImages,
				int				mapMode,
				int				xWinExt,
				int				yWinExt,
				int				twipsWide,
				int				twipsHigh );

extern void appMetaCleanDeviceContext(	DeviceContext *		dc );

extern int appMetaInitDeviceContext(
				DeviceContext *			dc,
				const PostScriptFontList *	psfl,
				int				objectCount,
				int				mapMode,
				int				xWinExt,
				int				yWinExt,
				int				xViewExt,
				int				yViewExt,
				int				twipsWide,
				int				twipsHigh );

extern void appMetaCleanObject( MetaFileObject *	mfo );

extern int appMeta_ExtTextOut(		SimpleInputStream *	sis,
					int			recordSize,
					int *			pX0,
					int *			pY0,
					int *			pCount,
					int *			pStyle,
					int *			pX1,
					int *			pY1,
					int *			pH1,
					int *			pW1,
					DeviceContext *		dc );

extern int appMeta_TextOut(	SimpleInputStream *	sis,
				int			recordSize,
				int *			pX0,
				int *			pY0,
				int *			pCount,
				DeviceContext *		dc );

extern int appMeta_GetCounts(	SimpleInputStream *	sis,
				int			recordSize,
				int *			pCount,
				int **			pCounts,
				DeviceContext *		dc );

extern int appMetaCreateBrushIndirect(	DeviceContext *		dc,
					int			recordSize,
					SimpleInputStream *	sis );

extern int appMetaCreatePenIndirect(	DeviceContext *		dc,
					int			recordSize,
					SimpleInputStream *	sis );

extern int appMetaCreateFontIndirect(	DeviceContext *		dc,
					int			recordSize,
					SimpleInputStream *	sis );

extern int appMetaCreatePatternBrush(	DeviceContext *		dc,
					int			recordSize,
					SimpleInputStream *	sis );

extern int appMetaCreatePalette(	DeviceContext *		dc,
					int			recordSize,
					SimpleInputStream *	sis );

extern int appMetaGetColor(		SimpleInputStream *	sis,
					RGB8Color *		rgb8 );

extern int appMetaListFontsPs(	PostScriptTypeList *	psfl,
				SimpleInputStream *	sis,
				const PostScriptFontList * x_psfl,
				const char *		prefix,
				int			mapMode,
				int			xExt,
				int			yExt,
				int			twipsWide,
				int			twipsHigh );

extern int appMetaBitmapImage(		AppBitmapImage **	pAbi,
					SimpleInputStream *	sis,
					int			expectBytes );

extern int appMetaSetTextAlign(		DeviceContext *		dc,
					int			recordSize,
					SimpleInputStream *	sis );

extern int appMetaDefineProcsetPs(	SimpleOutputStream *	sos );

extern int appMetaPlayWmfImg(	BitmapDescription *	pBd,
				unsigned char **	pBuffer,
				SimpleInputStream *	sis );

extern int appMetaSetMapMode(		DeviceContext *		dc,
					int			recordSize,
					SimpleInputStream *	sis );

extern int appMetaOffsetWindowOrg(	DeviceContext *		dc,
					int			recordSize,
					SimpleInputStream *	sis );

extern int appMetaSetWindowOrg(		DeviceContext *		dc,
					int			recordSize,
					SimpleInputStream *	sis );

extern int appMetaSetWindowExt(		DeviceContext *		dc,
					int			recordSize,
					SimpleInputStream *	sis );

extern int appMetaSetViewportOrg(	DeviceContext *		dc,
					int			recordSize,
					SimpleInputStream *	sis );

extern int appMetaSetViewportExt(	DeviceContext *		dc,
					int			recordSize,
					SimpleInputStream *	sis );

extern int appMetaScaleViewportExt(	DeviceContext *		dc,
					int			recordSize,
					SimpleInputStream *	sis );

extern int appMetaScaleWindowExt(	DeviceContext *		dc,
					int			recordSize,
					SimpleInputStream *	sis );

extern int appMetaSaveDC(		DeviceContext *		dc,
					int			recordSize,
					SimpleInputStream *	sis );

extern int appMetaRestoreDC(		DeviceContext *		dc,
					int			recordSize,
					SimpleInputStream *	sis );

extern int appMetaOffsetClipRgn(	DeviceContext *		dc,
					int			recordSize,
					SimpleInputStream *	sis );

extern int appMetaExcludeClipRect(	DeviceContext *		dc,
					int			recordSize,
					SimpleInputStream *	sis );

extern int appMetaSelectClipRgn(	DeviceContext *		dc,
					int			recordSize,
					SimpleInputStream *	sis );

extern int appMetaIntersectClipRect(	DeviceContext *		dc,
					int			recordSize,
					SimpleInputStream *	sis );

extern int appMetaSetBkColor(		DeviceContext *		dc,
					int			recordSize,
					SimpleInputStream *	sis );

extern int appMetaSetTextColor(		DeviceContext *		dc,
					int			recordSize,
					SimpleInputStream *	sis );

extern int appWinMetaGetPoints(		DeviceContext *		dc,
					int			count,
					SimpleInputStream *	sis );

extern void appMetaX11SetWindowsLineStyle(
					AppDrawingData *	add,
					int *			pDraw,
					int			ps,
					int			lpW,
					int			width );

