#   include	"appUtilConfig.h"

#   include	<stdio.h>
#   include	<stdlib.h>
#   include	<string.h>
#   include	<ctype.h>

#   include	"appSystem.h"
#   include	"utilX11FontNames.h"
#   include	"utilT1Entries.h"
#   include	"utilFontmap.h"

#   include	<appDebugon.h>

/************************************************************************/
/*									*/
/*  Get a line of input. (Essentially fgets())				*/
/*									*/
/************************************************************************/

static char *	utilT1GetInputLine(	char *		s,
					int *		pLen,
					int		size,
					FILE *		stream )
    {
    s= fgets( s, size, stream );

    if  ( s )
	{
	int		lineLength= strlen( s );

	if  ( lineLength > 0		&&
	      s[lineLength-1] == '\n'	)
	    { s[--lineLength]= '\0'; 	}

#	if 0
	No! are unix specific files

	if  ( lineLength > 0		&&
	      s[lineLength-1] == '\r'	)
	    { s[--lineLength]= '\0';	}

	/*  3  */
	if  ( lineLength > 0		&&
	      s[lineLength-1] == '\r'	)
	    { s[--lineLength]= '\0';	}
#	endif

	*pLen= lineLength;
	}

    return s;
    }

/************************************************************************/
/*									*/
/*  Read a fonts.dir file. for every font file name, X11 font name	*/
/*  combination, derive the X11 encoding from the name and remember	*/
/*  the X11 name with the encoding.					*/
/*									*/
/************************************************************************/

int utilSetT1EntriesForFonts(	const char *			t1MapFile,
				int				mapNames,
				const PostScriptFontList *	psfl )
    {
    int		rval= 0;

    FILE *	xfontDir= fopen( t1MapFile, "r" );
    char	scratch[256+1];
    int		lineLength;

    int		line= 0;

    if  ( ! xfontDir )
	{ SXDEB(t1MapFile,xfontDir); return -1;	}

    while( utilT1GetInputLine( scratch, &lineLength,
					    sizeof(scratch)- 1, xfontDir ) )
	{
	int			x11Start;
	int			x11Length;
	const char *		x11Name;
	int			res;

	int			fam;
	int			face;

	res= utilX11SplitFontsDirLine( scratch,
					&x11Start, &x11Length, line++ );
	if  ( res < 0 )
	    { SDEB(scratch); rval= -1; continue;	}
	if  ( res > 0 )
	    { continue;	}

	x11Name= scratch+ x11Start;

	for ( fam= 0; fam < psfl->psflFamilyCount; fam++ )
	    {
	    AfmFontInfo **		pafi;
	    const PsFontFamily *	aff= psfl->psflFamilies[fam];

	    pafi= aff->affFaces;
	    for ( face= 0; face < aff->affFaceCount; pafi++, face++ )
		{
		AfmFontInfo *		afi= *pafi;
		const char *		fontName;
		const char *		alias;

		fontName= alias= afi->afiFontName;

		if  ( mapNames )
		    {
		    alias= utilFontmapGetEntry( fontName );
		    if  ( ! alias )
			{ alias= fontName;	}
		    }

		if  ( ! strcmp( alias, scratch )		&&
		      psAddX11FontToInfo( afi, x11Name )	)
		    { LDEB(x11Length); rval= -1; goto ready;	}
		}
	    }
	}

  ready:
    if  ( xfontDir )
	{ fclose( xfontDir );	}

    return rval;
    }

