#   include	"bitmapConfig.h"

#   include	<stdio.h>
#   include	<stdlib.h>
#   include	<string.h>

#   include	"bitmap.h"

#   include	<appDebugon.h>

int bmSelect(	BitmapDescription *		bdOut,
		unsigned char **		pBuffer,
		const BitmapDescription *	bdIn,
		const unsigned char *		bufIn,
		int				x0,
		int				y0,
		int				wide,
		int				high )
    {
    BitmapDescription	bd;
    int			bytesPerRow;
    int			originBit;
    int			originByte;
    unsigned char *	buffer;
    int			row;

    if  ( x0+ wide > bdIn->bdPixelsWide	||
	  y0+ high > bdIn->bdPixelsHigh )
	{
	LLLLDEB(x0,y0,wide,high);
	LLDEB(bdIn->bdPixelsWide,bdIn->bdPixelsHigh);
	return -1;
	}

    /************************************************************/
    /*  Derive properties of output bitmap from input.		*/
    /************************************************************/
    bmCopyDescription( &bd, bdIn );
    bd.bdPixelsHigh= high;
    bd.bdPixelsWide= wide;
    bytesPerRow= ( bd.bdBitsPerPixel* bd.bdPixelsWide+ 7 )/ 8;
    bd.bdBytesPerRow= bytesPerRow;
    bd.bdBufferLength= high* bytesPerRow;

    /************************************************************/
    /*  How many bits to shift?					*/
    /************************************************************/
    originBit=  ( x0* bdIn->bdBitsPerPixel )% 8;
    originByte= ( x0* bdIn->bdBitsPerPixel )/ 8;

    /************************************************************/
    /*  Allocate new buffer.					*/
    /************************************************************/
    buffer= (unsigned char *)malloc( bd.bdBufferLength );
    if  ( ! buffer )
	{ LLDEB(bd.bdBufferLength,buffer); return -1;	}

    if  ( originBit == 0 )
	{
	for ( row= 0; row < high; row++ )
	    {
	    memcpy( buffer+ row* bytesPerRow, bufIn+
		    ( row+ y0 )* bdIn->bdBytesPerRow+ originByte, bytesPerRow );
	    }
	}
    else{
	int	rightShift= 8- originBit;

	for ( row= 0; row < high; row++ )
	    {
	    int				byte;
	    unsigned char *		to;
	    const unsigned char *	from;

	    to= buffer+ row* bytesPerRow;
	    from= bufIn+ ( row+ y0 )* bdIn->bdBytesPerRow+ originByte;

	    *to=   *from << originBit; from++;

	    for ( byte= 1; byte < bytesPerRow; byte++ )
		{
		*to |= *from >> rightShift; to++;
		*to=   *from << originBit; from++;
		}

	    *to |= *from >> rightShift;
	    }
	}

    *pBuffer= buffer;
    *bdOut= bd;

    return 0;
    }

/************************************************************************/
/*									*/
/*  Copy parts of an image into another one.				*/
/*									*/
/************************************************************************/

int bmCopyArea(			int				x0,
				int				y0,
				unsigned char *			buffer1,
				const unsigned char *		buffer2,
				const BitmapDescription *	bd1,
				const BitmapDescription *	bd2 )
    {
    int			x1;
    int			y1;

    int			y;

    int			targetBit;
    int			targetByte;
    int			copyBits;
    int			copyBytes;
    int			tailBits;

    /************************************************************/
    /*  Same kind of image?					*/
    /************************************************************/
    if  ( bd1->bdBitsPerSample != bd2->bdBitsPerSample )
	{ LLDEB(bd1->bdBitsPerSample,bd2->bdBitsPerSample); return -1;	}
    if  ( bd1->bdSamplesPerPixel != bd2->bdSamplesPerPixel )
	{ LLDEB(bd1->bdSamplesPerPixel,bd2->bdSamplesPerPixel); return -1; }
    if  ( bd1->bdBitsPerPixel != bd2->bdBitsPerPixel )
	{ LLDEB(bd1->bdBitsPerPixel,bd2->bdBitsPerPixel); return -1;	}
    if  ( bd1->bdColorEncoding != bd2->bdColorEncoding )
	{ LLDEB(bd1->bdColorEncoding,bd2->bdColorEncoding); return -1; }

    /************************************************************/
    /*  Origin inside target?					*/
    /************************************************************/
    if  ( x0 < 0 || y0 < 0 )
	{ LLDEB(x0,y0); return -1;	}
    if  ( x0 >= bd1->bdPixelsWide || y0 >= bd1->bdPixelsHigh )
	{ LLLLDEB(x0,y0,bd1->bdPixelsWide,bd1->bdPixelsHigh); return -1; }

    x1= x0+ bd2->bdPixelsWide;
    if  ( x1 > bd1->bdPixelsWide )
	{ x1=  bd1->bdPixelsWide;	}

    y1= y0+ bd2->bdPixelsHigh;
    if  ( y1 > bd1->bdPixelsHigh )
	{ y1=  bd1->bdPixelsHigh;	}

    /************************************************************/
    /*  How many bits to copy/shift?				*/
    /************************************************************/
    targetBit=  ( x0* bd1->bdBitsPerPixel )% 8;
    targetByte= ( x0* bd1->bdBitsPerPixel )/ 8;

    copyBits= ( x1- x0 )* bd1->bdBitsPerPixel;
    if  ( targetBit != 0 )
	{ copyBits -= 8- targetBit;	}
    copyBytes= copyBits/ 8;
    tailBits=  copyBits% 8;

    if  ( targetBit == 0 )
	{
	unsigned char	t1Mask= 0xff >> tailBits;
	unsigned char	t2Mask= 0xff << ( 8- tailBits );

	for ( y= y0; y < y1; y++ )
	    {
	    unsigned char *		r1;
	    const unsigned char *	r2;

	    r1= buffer1+ ( y-  0 )* bd1->bdBytesPerRow;
	    r2= buffer2+ ( y- y0 )* bd2->bdBytesPerRow;

	    r1 += targetByte;

	    memcpy( r1, r2, copyBytes );

	    if  ( tailBits > 0 )
		{
		r1 += copyBytes;
		r2 += copyBytes;

		*r1 &= t1Mask;
		*r1 |= *r2 & t2Mask;
		}
	    }

	return 0;
	}
    else{
	unsigned char	h1Mask= ~( 0xff >> targetBit );

	unsigned char	t1Mask= 0xff >> tailBits;
	unsigned char	t2Mask= 0xff << ( 8- tailBits );

	for ( y= y0; y < y1; y++ )
	    {
	    unsigned char *		r1;
	    const unsigned char *	r2;
	    int				i;
	    unsigned char		c;


	    r1= buffer1+ ( y-  0 )* bd1->bdBytesPerRow;
	    r2= buffer2+ ( y- y0 )* bd2->bdBytesPerRow;

	    r1 += targetByte;

	    c= ( *r1 & h1Mask ) | ( *r2 >> targetBit );
	    *(r1++)= c;

	    for ( i= 0; i < copyBytes; i++ )
		{
		c= *(r2++) << ( 8- targetBit );
		c |= *r2 >> targetBit;
		*(r1++)= c;
		}

	    if  ( tailBits > 0 )
		{
		c= *(r2++) << ( 8- targetBit );
		c |= *r2 >> targetBit;

		*r1 &= t1Mask;
		*r1 |= c & t2Mask;
		}
	    }

	return 0;
	}
    }

/************************************************************************/
/*									*/
/*  Transfer parts of an image into another one: Only locations that	*/
/*  are not transparent are affected.					*/
/*									*/
/************************************************************************/

static int bmPaintArea8(	int				x0,
				int				x1,
				int				y0,
				int				y1,
				unsigned char *			buffer1,
				const unsigned char *		buffer2,
				const BitmapDescription *	bd1,
				const BitmapDescription *	bd2 )
    {
    int		yf;
    int		yt;
    const int	spp1= bd1->bdSamplesPerPixel;
    const int	spp2= bd2->bdSamplesPerPixel;

    yf= 0;
    for ( yt= y0; yt < y1; yf++, yt++ )
	{
	int			x;
	unsigned char *		to= buffer1+ yt* bd1->bdBytesPerRow+ x0* spp1;
	const unsigned char *	fr= buffer2+ yf* bd2->bdBytesPerRow;

	for ( x= x0; x < x1; x++ )
	    {
	    if  ( fr[spp2- 1] )
		{ memcpy( to, fr, spp1 );	}

	    to += spp1; fr += spp2;
	    }
	}

    return 0;
    }

int bmPaintArea(		int				x0,
				int				y0,
				unsigned char *			buffer1,
				const unsigned char *		buffer2,
				const BitmapDescription *	bd1,
				const BitmapDescription *	bd2 )
    {
    int			x1;
    int			y1;

    if  ( ! bd2->bdHasAlpha )
	{ return bmCopyArea( x0, y0, buffer1, buffer2, bd1, bd2 );	}

    /************************************************************/
    /*  Same kind of image?					*/
    /************************************************************/
    if  ( bd1->bdBitsPerSample != bd2->bdBitsPerSample )
	{ LLDEB(bd1->bdBitsPerSample,bd2->bdBitsPerSample); return -1;	}
    if  ( bd1->bdColorEncoding != bd2->bdColorEncoding )
	{ LLDEB(bd1->bdColorEncoding,bd2->bdColorEncoding); return -1; }

    if  ( bd1->bdSamplesPerPixel+ !bd1->bdHasAlpha != bd2->bdSamplesPerPixel )
	{
	LLLDEB(bd1->bdSamplesPerPixel,bd1->bdHasAlpha,bd2->bdSamplesPerPixel);
	return -1;
	}

    /************************************************************/
    /*  Origin inside target?					*/
    /************************************************************/
    if  ( x0 < 0 || y0 < 0 )
	{ LLDEB(x0,y0); return -1;	}
    if  ( x0 >= bd1->bdPixelsWide || y0 >= bd1->bdPixelsHigh )
	{ LLLLDEB(x0,y0,bd1->bdPixelsWide,bd1->bdPixelsHigh); return -1; }

    x1= x0+ bd2->bdPixelsWide;
    if  ( x1 > bd1->bdPixelsWide )
	{ x1=  bd1->bdPixelsWide;	}

    y1= y0+ bd2->bdPixelsHigh;
    if  ( y1 > bd1->bdPixelsHigh )
	{ y1=  bd1->bdPixelsHigh;	}

    switch( bd1->bdBitsPerSample )
	{
	case 8:
	    return bmPaintArea8( x0, x1, y0, y1, buffer1, buffer2, bd1, bd2 );

	default:
	    LDEB(bd1->bdBitsPerSample); return -1;
	}
    }

