/************************************************************************/
/*									*/
/*  Manage a list of fields.						*/
/*									*/
/************************************************************************/

#   include	"docBufConfig.h"

#   include	<stdlib.h>
#   include	<string.h>
#   include	<stdio.h>
#   include	<limits.h>

#   include	<appDebugon.h>

#   include	"docDocumentFieldList.h"

/************************************************************************/
/*									*/
/*  Manage a list of fields.						*/
/*									*/
/************************************************************************/

/************************************************************************/
/*									*/
/*  Make a copy of a field.						*/
/*									*/
/************************************************************************/

DocumentField *	docClaimFieldCopy(	DocumentFieldList *	dfl,
					const DocumentField *	dfFrom,
					const SelectionScope *	ss,
					const EditPosition *	epStart )
    {
    DocumentField *	dfTo;

    dfTo= docClaimField( dfl );
    if  ( ! dfTo )
	{ XDEB(dfTo); return (DocumentField *)0;	}

    if  ( utilCopyMemoryBuffer( &(dfTo->dfData), &(dfFrom->dfData) ) 	||
	  utilCopyMemoryBuffer( &(dfTo->dfInstructions),
					&(dfFrom->dfInstructions) )	)
	{
	docDeleteFieldFromList( dfl, dfTo );
	return (DocumentField *)0;
	}

    dfTo->dfPage= -1;
    dfTo->dfKind= dfFrom->dfKind;

    dfTo->dfSelectionScope= *ss;
    dfTo->dfHeadPosition= *epStart;

    return dfTo;
    }

DocumentField * docGetFieldByNumber(	const DocumentFieldList *	dfl,
					int				n )
    {
    int			page= n/ DFpageSIZE;
    DocumentField *	fieldPage;

    if  ( page >= dfl->dflFieldPageCount )
	{ return (DocumentField *)0;	}

    fieldPage= dfl->dflFieldPages[page];
    if  ( ! fieldPage )
	{ return (DocumentField *)0;	}

    if  ( fieldPage[n%DFpageSIZE].dfFieldNumber != n )
	{ return (DocumentField *)0;	}

    return &(fieldPage[n%DFpageSIZE]);
    }

int docGetFieldKindByNumber(	const DocumentFieldList *	dfl,
				int				n )
    {
    DocumentField *	df= docGetFieldByNumber( dfl, n );

    if  ( ! df )
	{ return -1;	}

    return df->dfKind;
    }

void docDeleteFieldFromList(	DocumentFieldList *		dfl,
				DocumentField *			df )
    {
    int		n= df->dfFieldNumber;

    if  ( docGetFieldByNumber( dfl, n ) != df )
	{ LDEB(n);	}

    docCleanDocumentField( df );
    docInitDocumentField( df );

    if  ( dfl->dflFreeField == n )
	{ LLDEB(dfl->dflFreeField,n); return;	}

    df->dfFieldNumber= dfl->dflFreeField;
    dfl->dflFreeField= n;

    return;
    }

DocumentField * docClaimField(	DocumentFieldList *	dfl )
    {
    int			n;
    int			page;
    DocumentField *	fieldPage;
    DocumentField *	df;

    if  ( dfl->dflFreeField >= 0 )
	{
	n= dfl->dflFreeField;
	page= n/ DFpageSIZE;
	fieldPage= dfl->dflFieldPages[page];

	df= &(fieldPage[n%DFpageSIZE]);
	dfl->dflFreeField= df->dfFieldNumber;

	df->dfFieldNumber= n;
	return df;
	}

    n= dfl->dflFieldCount;
    page= n/ DFpageSIZE;

    if  ( page >= dfl->dflFieldPageCount )
	{
	int	newPageCount= page+ 1;

	DocumentField ** fresh= realloc( dfl->dflFieldPages,
				    newPageCount* sizeof(DocumentField *) );

	if  ( ! fresh )
	    { LXDEB(n,fresh); return (DocumentField *)0;	}
	dfl->dflFieldPages= fresh;

	while( dfl->dflFieldPageCount < newPageCount )
	    { fresh[dfl->dflFieldPageCount++]= (DocumentField *)0;	}
	}

    fieldPage= dfl->dflFieldPages[page];
    if  ( ! fieldPage )
	{
	int		i;

	dfl->dflFieldPages[page]= fieldPage=
				malloc( DFpageSIZE* sizeof(DocumentField) );
	if  ( ! fieldPage )
	    { XDEB(fieldPage); return (DocumentField *)0;	}

	for ( i= 0; i < DFpageSIZE; i++ )
	    { docInitDocumentField( fieldPage+ i );	}
	}

    if  ( dfl->dflFieldCount <= n )
	{ dfl->dflFieldCount= n+ 1;	}

    fieldPage= dfl->dflFieldPages[page];
    df= &(fieldPage[n%DFpageSIZE]);
    df->dfFieldNumber= n;
    return df;
    }

void docInitFieldList(	DocumentFieldList *	dfl )
    {
    dfl->dflFieldCount= 0;
    dfl->dflFieldPages= (DocumentField **)0;
    dfl->dflFieldPageCount= 0;
    dfl->dflFreeField= -1;

    return;
    }

void docCleanFieldList(	DocumentFieldList *	dfl )
    {
    int		page;
    int		i;

    for ( page = 0; page < dfl->dflFieldPageCount; page++ )
	{
	DocumentField *	fieldPage= dfl->dflFieldPages[page];

	if  ( fieldPage )
	    {
	    for ( i= 0; i < DFpageSIZE; i++ )
		{ docCleanDocumentField( fieldPage+ i );	}

	    free( fieldPage );
	    }
	}

    if  ( dfl->dflFieldPages )
	{ free( dfl->dflFieldPages );	}

    return;
    }

/************************************************************************/
/*									*/
/*  Remember the page where a field occurs.				*/
/*									*/
/************************************************************************/

void docSetPageOfField(		DocumentFieldList *	dfl,
				int			n,
				int			page )
    {
    DocumentField *	df= docGetFieldByNumber( dfl, n );

    if  ( ! df )
	{ LPDEB(n,df); return;	}

    df->dfPage= page;

    return;
    }

/************************************************************************/
/*									*/
/*  Look for a bookmark in the field list.				*/
/*									*/
/************************************************************************/

int docGetBookmarkField(	DocumentField **		pDf,
				const DocumentFieldList *	dfl,
				const char *			markName,
				int				markSize )
    {
    int			n;

    for ( n= 0; n < dfl->dflFieldCount; n++ )
	{
	DocumentField *	df= docGetFieldByNumber( dfl, n );

	if  ( ! df )
	    { continue;	}

	if  ( docFieldMatchesBookmark( df, markName, markSize )	)
	    { *pDf= df; return n;	}
	}

    return -1;
    }

/************************************************************************/
/*									*/
/*  Build a list of bookmarks to be used in a GUI.			*/
/*									*/
/************************************************************************/

static int docCompareBookmarkStrings(	const void *	pv1,
					const void *	pv2 )
    {
    const char * const *	ps1= (const char * const *)pv1;
    const char * const *	ps2= (const char * const *)pv2;

    return strcmp( *ps1, *ps2 );
    }

int docMakeBookmarkList(	char ***			pBookmarks,
				int *				pBookmarkCount,
				const DocumentFieldList *	dfl )
    {
    int				rval= 0;
    int				n;

    int				bookmarkCount= 0;
    char **			bookmarks= (char **)0;

    if  ( dfl->dflFieldCount == 0 )
	{ goto ready;	}

    /*  2  */
    bookmarks= (char **)malloc( dfl->dflFieldCount* sizeof(char *) );
    if  ( ! bookmarks )
	{ LXDEB(dfl->dflFieldCount,bookmarks); rval= -1; goto ready;	}

    for ( n= 0; n < dfl->dflFieldCount; n++ )
	{
	DocumentField *	df= docGetFieldByNumber( dfl, n );

	const char *	name;
	int		size;

	if  ( ! df )
	    { continue;	}

	if  ( df->dfKind != DOCfkBOOKMARK		||
	      docFieldGetBookmark( df, &name, &size )	)
	    { continue;		}

	if  ( size == 0			||
	      name[0] == '_'		||
	      size > DOCmaxBOOKMARK	)
	    { continue;	}

	bookmarks[bookmarkCount]= malloc( size+ 1 );
	if  ( ! bookmarks[bookmarkCount] )
	    { LXDEB(size,bookmarks[bookmarkCount]); rval= -1; goto ready; }

	strncpy( bookmarks[bookmarkCount], name, size )[size]= '\0';
	bookmarkCount++;
	}

    /*  3  */
    qsort( bookmarks, bookmarkCount, sizeof(char *),
					    docCompareBookmarkStrings );

    *pBookmarks= bookmarks; bookmarks= (char **)0; /* steal */
    *pBookmarkCount= bookmarkCount; bookmarkCount= 0;

  ready:

    for ( n= 0; n < bookmarkCount; n++ )
	{
	if  ( bookmarks[n] )
	    { free( bookmarks[n] );	}
	}

    if  ( bookmarks )
	{ free( bookmarks );	}

    return rval;
    }

